"""SOAP Client implementation."""

from __future__ import annotations

import logging
import socket
import time
import traceback
from http.client import (
    HTTPConnection,
    HTTPException,
    HTTPResponse,
    HTTPSConnection,
    NotConnected,
    UnknownTransferEncoding,
)
from threading import Lock
from typing import TYPE_CHECKING, Protocol

from lxml import etree

from sdc11073 import commlog, observableproperties
from sdc11073.httpserver.compression import CompressionHandler
from sdc11073.httpserver.httpreader import HTTPReader, mk_chunks
from sdc11073.namespaces import default_ns_helper as ns_hlp
from sdc11073.pysoap.soapenvelope import Fault

if TYPE_CHECKING:
    from collections.abc import Iterable
    from ssl import SSLContext

    from sdc11073.consumer.manipulator import RequestManipulatorProtocol
    from sdc11073.definitions_base import BaseDefinitions
    from sdc11073.loghelper import LoggerAdapter
    from sdc11073.pysoap.msgfactory import CreatedMessage
    from sdc11073.pysoap.msgreader import MessageReader, ReceivedMessage


class HTTPConnectionNoDelay(HTTPConnection):
    """Connect method sets specific socket options."""

    def connect(self):
        """Connect to a host on a given port."""
        super().connect()
        self.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, True)


class HTTPSConnectionNoDelay(HTTPSConnection):
    """Connect method sets specific socket options."""

    def connect(self):
        """Connect to a host on a given SSL port."""
        super().connect()
        self.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, True)


class HTTPReturnCodeError(HTTPException):
    """HTTPReturnCodeError is used to map http return codes to Python exceptions."""

    def __init__(self, status: int, reason: str, soap_fault: Fault | None):
        super().__init__()
        self.status = status
        self.reason = reason
        self.soap_fault = soap_fault

    def __repr__(self) -> str:
        return f'HTTPReturnCodeError(status={self.status}, reason={self.reason} fault={self.soap_fault}'


class SoapClientProtocol(Protocol):
    """The expected interface of a soap client."""

    sock_name: tuple[str, int] | None

    def __init__(  # noqa: PLR0913
        self,
        netloc: str,
        socket_timeout: float,
        logger: LoggerAdapter,
        ssl_context: SSLContext | None,
        sdc_definitions: type[BaseDefinitions],
        msg_reader: MessageReader,
        supported_encodings: Iterable[str] | None = None,
        request_encodings: Iterable[str] | None = None,
        chunk_size: int = 0,
    ): ...

    def post_message_to(
        self,
        hosted_service_path: str,
        message: CreatedMessage,
        msg: str = '',
        request_manipulator: RequestManipulatorProtocol | None = None,
        validate: bool = True,
    ) -> ReceivedMessage | None:
        """Send the message and return None if the response is empty else the received response."""
        ...

    def get_from_url(self, url: str, msg: str) -> bytes:
        """Send a GET request and return content of response."""
        ...

    def is_closed(self) -> bool:
        """Return True if connection is closed."""

    def connect(self):
        """Connect to net location."""

    def close(self):
        """Close connection."""

    @property
    def sock(self) -> socket.SocketType | None:
        """Return used socket."""


class SoapClient:
    """SOAP Client wraps an http connection. It can send / receive SoapEnvelopes."""

    _used_soap_clients = 0

    roundtrip_time = observableproperties.ObservableProperty()

    def __init__(  # noqa: PLR0913
        self,
        netloc: str,
        socket_timeout: float,
        logger: LoggerAdapter,
        ssl_context: SSLContext | None,
        sdc_definitions: type[BaseDefinitions],
        msg_reader: MessageReader,
        supported_encodings: Iterable[str] | None = None,
        request_encodings: Iterable[str] | None = None,
        chunk_size: int = 0,
    ):
        """Connect to one url.

        :param netloc: the location of the service (domain name:port) ###url of the service
        :param logger: a python logger instance
        :param ssl_context: an optional sll.SSLContext instance
        :param sdc_definitions: needed to normalize and de-normalize xml text.
        :param supported_encodings: Configured set of encodings that can be used.
                                    If None, all available encodings are used.
                                    This used for decompression of received responses.
                                    If this is an empty list, no compression is supported.
        :param request_encodings: An optional list of encodings that the other side accepts.
                                  It is used to compress requests.
                                  If not set, requests will not be compressed.
                                  If set, then the http request will be compressed using this method
        :param chunk_size: if value > 0, message is split into chunks of this size.
        """
        self._log = logger
        self._ssl_context = ssl_context
        self._sdc_definitions = sdc_definitions
        self._msg_reader = msg_reader
        self._netloc = netloc
        self._socket_timeout = socket_timeout
        self._http_connection = None  # connect later on demand
        self.__class__._used_soap_clients += 1  # noqa: SLF001
        self._client_number = self.__class__._used_soap_clients  # noqa: SLF001
        self._log.info('created soap client No. {} for {}', self._client_number, netloc)
        self.supported_encodings = (
            supported_encodings if supported_encodings is not None else CompressionHandler.available_encodings
        )
        # request_encodings contains the compression algorithms that the other side accepts ( set at runtime)
        self.request_encodings = request_encodings if request_encodings is not None else []
        self._get_headers = self._make_get_headers()
        self._lock = Lock()
        self._chunk_size = chunk_size
        self._has_connection_error = False  # used to avoid implicit connects after an error
        self.sock_name: tuple[str, int] | None = None

    @property
    def netloc(self) -> str:
        """Return location, e.g.127.0.0.1:9999."""
        return self._netloc

    @property
    def sock(self) -> socket.SocketType | None:
        """Return used socket."""
        return None if self._http_connection is None else self._http_connection.sock

    def _mk_http_connection(self) -> [HTTPSConnectionNoDelay, HTTPConnectionNoDelay]:
        """Establish connection.

        Soap client never sends very large requests, the largest packages are notifications.
        We can use TCP_NODELAY for a little faster transmission.
        """
        if self._ssl_context is not None:
            conn = HTTPSConnectionNoDelay(self._netloc, context=self._ssl_context, timeout=self._socket_timeout)
        else:
            conn = HTTPConnectionNoDelay(self._netloc, timeout=self._socket_timeout)
        return conn

    def connect(self):
        """Connect to netloc."""
        self._has_connection_error = False
        self._http_connection = self._mk_http_connection()
        self._http_connection.connect()  # connect now so that we have own address and port for logging
        self.sock_name = self._http_connection.sock.getsockname()
        self._log.info(
            'soap client No. {} uses connection={}:{}',
            self._client_number,
            self.sock_name[0],
            self.sock_name[1],
        )

    def close(self):
        """Close connection."""
        with self._lock:
            self._close_without_lock()

    def _close_without_lock(self):
        self.sock_name = None
        if self._http_connection is not None:
            self._log.info('closing soapClientNo {} for {}', self._client_number, self._netloc)
            self._http_connection.close()
            self._http_connection = None

    def is_closed(self) -> bool:
        """Return True if connection is closed."""
        return self._http_connection is None or self._http_connection.sock is None

    @staticmethod
    def _prepare_message(
        created_message: CreatedMessage,
        request_manipulator: RequestManipulatorProtocol | None,
        validate: bool,
    ) -> bytes:
        if hasattr(request_manipulator, 'manipulate_soapenvelope'):
            tmp = request_manipulator.manipulate_soapenvelope(created_message.p_msg)
            if tmp:
                created_message.p_msg = tmp
                # in this case do not validate , because the manipulator might intentionally have created invalid xml.
                validate = False
        xml_request = created_message.serialize(request_manipulator=request_manipulator, validate=validate)

        if hasattr(request_manipulator, 'manipulate_string'):
            tmp = request_manipulator.manipulate_string(xml_request)
            if tmp:
                xml_request = tmp
        return xml_request

    def post_message_to(
        self,
        path: str,
        created_message: CreatedMessage,
        msg: str = '',
        request_manipulator: RequestManipulatorProtocol | None = None,
        validate: bool = True,
    ) -> ReceivedMessage | None:
        """Post created message to netloc/path.

        :param path: url path component
        :param created_message: The message that shall be sent
        :param msg: used in logs, helps to identify the context in which the method was called
        :param request_manipulator: see documentation of RequestManipulatorProtocol
        :param validate: set to False if no schema validation shall be done
        """
        if self.is_closed() and not self._has_connection_error:
            # implicit connect
            self.connect()
        if self.is_closed():
            raise NotConnected
        xml_request = self._prepare_message(created_message, request_manipulator, validate)
        started = time.perf_counter()
        try:
            with self._lock:
                http_response, xml_response = self._send_soap_request(path, xml_request, msg)
        finally:
            self.roundtrip_time = time.perf_counter() - started  # set roundtrip time even if method raises an exception
        if not xml_response:  # empty response
            return None

        message_data = self._msg_reader.read_received_message(xml_response)
        if message_data.action == f'{ns_hlp.WSA.namespace}/fault':
            soap_fault = Fault.from_node(message_data.p_msg.msg_node)
            raise HTTPReturnCodeError(http_response.status, http_response.reason, soap_fault)
        return message_data

    def _send_soap_request(self, path: str, xml: bytes, log_msg: str) -> tuple[HTTPResponse, bytes]:  # noqa: PLR0915, PLR0912, C901
        """Send SOAP request."""
        logging.getLogger(commlog.SOAP_REQUEST_OUT).debug(xml, extra={'http_method': 'POST'})
        self._log.debug("{}:POST to netloc='{}' path='{}'", log_msg, self._netloc, path)

        headers = {
            'Content-type': 'application/soap+xml; charset=utf-8',
            'user_agent': 'pysoap',
            'Connection': 'keep-alive',
        }
        # set accepted encodings
        if self.supported_encodings:
            headers['Accept-Encoding'] = ','.join(self.supported_encodings)
        # if possible encode ( compress) xml data
        if self.request_encodings:
            for compr in self.request_encodings:
                if compr in self.supported_encodings:
                    xml = CompressionHandler.compress_payload(compr, xml)
                    headers['Content-Encoding'] = compr
                    break
        # split message into chunks?
        if self._chunk_size > 0:
            headers['transfer-encoding'] = 'chunked'
            xml = mk_chunks(xml, self._chunk_size)
        else:
            headers['Content-Length'] = str(len(xml))

        # send the request
        try:
            self._http_connection.request('POST', path, body=xml, headers=headers)
        except HTTPException as ex:
            self._log.warn('{}: could not send request, http exception = {!r}', log_msg, ex)
            self._close_without_lock()
            self._has_connection_error = True
            raise NotConnected from ex
        except OSError as ex:
            if ex.errno in (10053, 10054):
                self._log.warn('{}: could not send request to {}, OSError={!r}', log_msg, self.netloc, ex)
            else:
                self._log.warn(
                    '{}: could not send request to {}, OSError={}',
                    log_msg,
                    self.netloc,
                    traceback.format_exc(),
                )
            self._has_connection_error = True
            self._close_without_lock()
            raise NotConnected from ex
        except Exception as ex:
            self._log.warn(
                "{}: POST to netloc='{}' path='{}': could not send request, error={!r}\n{}",
                log_msg,
                self._netloc,
                path,
                ex,
                traceback.format_exc(),
            )
            self._has_connection_error = True
            self._close_without_lock()
            raise NotConnected from ex

        # read the response
        try:
            response = self._http_connection.getresponse()
        except HTTPException as ex:
            self._log.warn('{}: could not receive response, http exception = {!r} ', log_msg, ex)
            self._has_connection_error = True
            self._close_without_lock()
            raise NotConnected from ex
        except OSError as ex:
            if ex.errno in (10053, 10054):
                self._log.warn('{}: could not receive response, OSError={!r}', log_msg, ex)
            else:
                self._log.warn(
                    '{}: could not receive response, OSError={!r} ({!r})\n{}',
                    log_msg,
                    ex.errno,
                    ex,
                    traceback.format_exc(),
                )
            self._has_connection_error = True
            self._close_without_lock()
            raise NotConnected from ex
        except Exception as ex:
            self._log.warn(
                "{}: POST to netloc='{}' path='{}': could not receive response, error={!r}\n{}",
                log_msg,
                self._netloc,
                path,
                ex,
                traceback.format_exc(),
            )
            self._has_connection_error = True
            self._close_without_lock()
            raise NotConnected from ex

        content = HTTPReader.read_response_body(response)

        if response.status >= 300:  # noqa: PLR2004
            self._log.error(
                "{}: POST to netloc='{}' path='{}': could not send request, HTTP response={}\ncontent='{}'",
                log_msg,
                self._netloc,
                path,
                response.status,
                content.decode('utf-8'),
            )
            try:
                tmp = self._msg_reader.read_received_message(content)
            except etree.XMLSyntaxError as ex:
                raise HTTPReturnCodeError(response.status, response.reason, None) from ex
            else:
                soap_fault = Fault.from_node(tmp.p_msg.msg_node)
                raise HTTPReturnCodeError(response.status, response.reason, soap_fault)

        response_headers = {k.lower(): v for k, v in response.getheaders()}

        self._log.debug('{}: response:{}; content has {} Bytes ', log_msg, response_headers, len(content))
        logging.getLogger(commlog.SOAP_RESPONSE_IN).debug(content, extra={'http_method': 'POST'})
        return response, content

    def _make_get_headers(self) -> dict[str, str]:
        headers = {
            'user_agent': 'pysoap',
            'Connection': 'keep-alive',
        }
        if self.supported_encodings:
            headers['Accept-Encoding'] = ', '.join(self.supported_encodings)
        return headers

    def get_from_url(self, url: str, msg: str) -> bytes:
        """Send a GET request and return content of response."""
        if self.is_closed() and not self._has_connection_error:
            # implicit connect
            self.connect()
        if self.is_closed():
            raise NotConnected

        if not url.startswith('/'):
            url = '/' + url
        self._log.debug('{} Get {}/{}', msg, self._netloc, url)
        with self._lock:
            self._http_connection.request('GET', url, headers=self._get_headers)
            response = self._http_connection.getresponse()
            headers = {k.lower(): v for k, v in response.getheaders()}
            _content = response.read()
            if 'content-encoding' in headers:
                enc = headers['content-encoding']
                if enc in self.supported_encodings:
                    content = CompressionHandler.decompress_payload(enc, _content)
                else:
                    self._log.warn('{}: unsupported compression ', headers['content-encoding'])
                    raise UnknownTransferEncoding
            else:
                content = _content
        return content




from __future__ import annotations

from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from collections.abc import Iterable

    from sdc11073.definitions_base import AbstractDataModel
    from sdc11073.xml_types.pm_types import Annotation, ComponentActivation


class RtSampleArray:
    """RtSampleArray contains a list of waveform values plus time stamps and annotations.

    It is the output of the waveform provider and contains all data that is needed to create waveform notifications.
    """

    def __init__(self, model: AbstractDataModel,
                 determination_time: float | None,
                 sample_period: float,
                 samples: list[float],
                 activation_state: ComponentActivation):
        """Construct a RtSampleArray.

        :param determination_time: the time stamp of the first value in samples, can be None if not active
        :param sample_period: the time difference between two samples
        :param samples: a list of 2-tuples (value (float or int), flag annotation_trigger)
        :param activation_state: one of pmtypes.ComponentActivation values
        """
        self._model = model
        self.determination_time = determination_time
        self.sample_period = sample_period
        self.samples = samples
        self.activation_state = activation_state
        self.annotations = []
        self.apply_annotations = []

    def _nearest_index(self, timestamp: float) -> int | None:
        """Find the realtime sample that's determination time is nearest to given timestamp."""
        if self.determination_time is None:  # when deactivated, determinationTime is None
            return None
        # check if timestamp is too old. Accept 0.5*sample period as tolerance.
        # self.determination_time is the timestamp of the first element in the array
        delta = self.sample_period * 0.5
        if timestamp < (self.determination_time - delta):
            return None  # timestamp too small (== to old)
        youngest_determination_time = self.determination_time + len(self.samples) * self.sample_period
        # check if timestamp is too young. Accept 0.5*sample period as tolerance.
        if timestamp >= youngest_determination_time + delta:
            return None  # timestamp too big (== to young)
        # calculate the position
        pos = (timestamp - self.determination_time) / self.sample_period
        return int(pos) + 1 if pos % 1 >= 0.5 else int(pos)  # return the nearest element

    def add_annotations_at(self, annotation: Annotation, timestamps: Iterable[float]):
        """Add annotation at the waveform samples nearest to timestamps."""
        applied = False
        annotation_index = len(self.annotations)  # Index is zero-based
        for timestamp in timestamps:
            i = self._nearest_index(timestamp)
            if i is not None:
                self.apply_annotations.append(self._model.pm_types.ApplyAnnotation(annotation_index, i))
                applied = True
        if applied:
            self.annotations.append(annotation)


class AnnotatorProtocol(Protocol):
    """An Annotator adds Annotations to waveforms. It mimics things like start of inspiration cycle etc."""

    annotation: Annotation
    trigger_handle: str
    annotated_handles: list[str]

    def __init__(self, annotation: Annotation, trigger_handle: str, annotated_handles: list[str]):
        """Construct an annotator."""

    def get_annotation_timestamps(self, rt_sample_array: RtSampleArray) -> list[float]:
        """Analyze the rt_sample_array and return timestamps for annotations.

        Return a list of timestamps, can be empty.
        """

class Annotator:
    """Annotator is a sample of how to apply annotations.

    This annotator triggers an annotation when the value changes from <= 0 to > 0.
    """

    def __init__(self, annotation: Annotation, trigger_handle: str, annotated_handles: list[str]):
        """Construct an annotator.

        :param annotation: Annotation
        :param trigger_handle: the handle of the state that triggers an annotation
        :param annotated_handles: list of handles that get annotated
        """
        self.annotation = annotation
        self.trigger_handle = trigger_handle
        self.annotated_handles = annotated_handles
        self._last_value = 0.0

    def get_annotation_timestamps(self, rt_sample_array: RtSampleArray) -> list[float]:
        """Analyze the rt_sample_array and return timestamps for annotations."""
        ret = []
        for i, rt_sample in enumerate(rt_sample_array.samples):
            if self._last_value <= 0 and rt_sample > 0:
                ret.append(rt_sample_array.determination_time + i * rt_sample_array.sample_period)
            self._last_value = rt_sample
        return ret



"""Implementation of some data types used in Message Model."""

# ruff: noqa: D101, D102

from __future__ import annotations

import enum
from abc import abstractmethod
from typing import TYPE_CHECKING, Any

from sdc11073.mdib.descriptorcontainers import (
    AbstractDescriptorContainer,
    AbstractMetricDescriptorContainer,
    ChannelDescriptorContainer,
    MdsDescriptorContainer,
    VmdDescriptorContainer,
)
from sdc11073.mdib.descriptorcontainers import (
    get_container_class as get_descriptor_container_class,
)
from sdc11073.mdib.statecontainers import (
    AbstractAlertStateContainer,
    AbstractContextStateContainer,
    AbstractDeviceComponentStateContainer,
    AbstractMetricStateContainer,
    AbstractOperationStateContainer,
    AbstractStateContainer,
    RealTimeSampleArrayMetricStateContainer,
)
from sdc11073.mdib.statecontainers import (
    get_container_class as get_state_container_class,
)
from sdc11073.namespaces import default_ns_helper

if TYPE_CHECKING:
    from sdc11073 import xml_utils
    from sdc11073.mdib.mdibbase import MdibVersionGroup

from sdc11073.xml_types import ext_qnames as ext
from sdc11073.xml_types import msg_qnames as msg
from sdc11073.xml_types import pm_qnames as pm
from sdc11073.xml_types import pm_types
from sdc11073.xml_types import xml_structure as cp
from sdc11073.xml_types.actions import Actions
from sdc11073.xml_types.basetypes import MessageType
from sdc11073.xml_types.dataconverters import UnsignedIntConverter
from sdc11073.xml_types.pm_types import (
    ContainmentTree,
    InstanceIdentifier,
    LocalizedText,
    LocalizedTextWidth,
    PropertyBasedPMType,
)


class StringEnum(str, enum.Enum):
    def __str__(self):
        return str(self.value)


class DescriptionModificationType(StringEnum):
    CREATE = 'Crt'
    UPDATE = 'Upt'
    DELETE = 'Del'


class InvocationState(StringEnum):
    WAIT = 'Wait'  # Wait = Waiting. The operation has been queued and waits for execution.
    START = 'Start'  # Start = Started. The execution of the operation has been started
    CANCELLED = 'Cnclld'  # Cnclld = Cancelled. The execution has been cancelled by the SERVICE PROVIDER.
    CANCELLED_MANUALLY = (
        'CnclldMan'  # CnclldMan = Cancelled Manually. The execution has been cancelled by the operator.
    )
    FINISHED = 'Fin'  # Fin = Finished. The execution has been finished.
    FINISHED_MOD = 'FinMod'  # FinMod = Finished with modification.
    # As the requested target value could not be reached, the next best value has been chosen
    # and used as target value.
    FAILED = 'Fail'  # The execution has been failed.


class InvocationError(StringEnum):
    UNSPECIFIED = 'Unspec'  # An unspecified error has occurred. No more information about the error is available.
    UNKNOWN_OPERATION = 'Unkn'  # Unknown Operation. The HANDLE to the operation object is not known.
    INVALID_VALUE = (
        'Inv'  # Invalid Value. The HANDLE to the operation object does not match the invocation request message
    )
    OTHER = 'Oth'  # Another type of error has occurred. More information on the error MAY be available.


### Reports ###
class AbstractReportPart(PropertyBasedPMType):
    SourceMds = cp.NodeStringProperty(msg.SourceMds, is_optional=True)
    _props = ('SourceMds',)


class AbstractReport(MessageType):
    MdibVersion = cp.IntegerAttributeProperty('MdibVersion', implied_py_value=0)
    SequenceId = cp.StringAttributeProperty('SequenceId')
    InstanceId = cp.IntegerAttributeProperty('InstanceId')
    _props = ('MdibVersion', 'SequenceId', 'InstanceId')
    additional_namespaces = (default_ns_helper.XSI,)

    def set_mdib_version_group(self, mdib_version_group: MdibVersionGroup) -> None:
        self.MdibVersion = mdib_version_group.mdib_version
        self.SequenceId = mdib_version_group.sequence_id
        self.InstanceId = mdib_version_group.instance_id


class ReportPartValuesList(AbstractReportPart):
    @property
    @abstractmethod
    def values_list(self) -> list[Any]:
        raise NotImplementedError


##### Metric Report ###
class MetricReportPart(ReportPartValuesList):
    MetricState = cp.ContainerListProperty(
        msg.MetricState,
        value_class=AbstractContextStateContainer,
        cls_getter=get_state_container_class,
        ns_helper=default_ns_helper,
    )
    _props = ('MetricState',)

    @property
    def values_list(self) -> list[Any]:
        return self.MetricState


class AbstractMetricReport(AbstractReport):
    ReportPart = cp.SubElementListProperty(msg.ReportPart, value_class=MetricReportPart)
    _props = ('ReportPart',)

    def add_report_part(self) -> MetricReportPart:
        self.ReportPart.append(MetricReportPart())
        return self.ReportPart[-1]


class EpisodicMetricReport(AbstractMetricReport):
    NODETYPE = msg.EpisodicMetricReport
    action = Actions.EpisodicMetricReport


class PeriodicMetricReport(AbstractMetricReport):
    NODETYPE = msg.PeriodicMetricReport
    action = Actions.PeriodicMetricReport


##### Context Report ###
class ContextReportPart(ReportPartValuesList):
    ContextState = cp.ContainerListProperty(
        msg.ContextState,
        value_class=AbstractContextStateContainer,
        cls_getter=get_state_container_class,
        ns_helper=default_ns_helper,
    )
    _props = ('ContextState',)

    @property
    def values_list(self) -> list[Any]:
        return self.ContextState


class AbstractContextReport(AbstractReport):
    ReportPart = cp.SubElementListProperty(msg.ReportPart, value_class=ContextReportPart)
    _props = ('ReportPart',)

    def add_report_part(self) -> ContextReportPart:
        self.ReportPart.append(ContextReportPart())
        return self.ReportPart[-1]


class EpisodicContextReport(AbstractContextReport):
    NODETYPE = msg.EpisodicContextReport
    action = Actions.EpisodicContextReport


class PeriodicContextReport(AbstractContextReport):
    NODETYPE = msg.PeriodicContextReport
    action = Actions.PeriodicContextReport


##### Operational State Report ###
class OperationalStateReportPart(ReportPartValuesList):
    OperationState = cp.ContainerListProperty(
        msg.OperationState,
        value_class=AbstractOperationStateContainer,
        cls_getter=get_state_container_class,
        ns_helper=default_ns_helper,
    )
    _props = ('OperationState',)

    @property
    def values_list(self) -> list[Any]:
        return self.OperationState


class AbstractOperationalStateReport(AbstractReport):
    ReportPart = cp.SubElementListProperty(msg.ReportPart, value_class=OperationalStateReportPart)
    _props = ('ReportPart',)

    def add_report_part(self) -> OperationalStateReportPart:
        self.ReportPart.append(OperationalStateReportPart())
        return self.ReportPart[-1]


class EpisodicOperationalStateReport(AbstractOperationalStateReport):
    NODETYPE = msg.EpisodicOperationalStateReport
    action = Actions.EpisodicOperationalStateReport


class PeriodicOperationalStateReport(AbstractOperationalStateReport):
    NODETYPE = msg.PeriodicOperationalStateReport
    action = Actions.PeriodicOperationalStateReport


##### Alert Report ###
class AlertReportPart(ReportPartValuesList):
    AlertState = cp.ContainerListProperty(
        msg.AlertState,
        value_class=AbstractAlertStateContainer,
        cls_getter=get_state_container_class,
        ns_helper=default_ns_helper,
    )
    _props = ('AlertState',)

    @property
    def values_list(self) -> list[Any]:
        return self.AlertState


class AbstractAlertReport(AbstractReport):
    ReportPart = cp.SubElementListProperty(msg.ReportPart, value_class=AlertReportPart)
    _props = ('ReportPart',)

    def add_report_part(self) -> AlertReportPart:
        self.ReportPart.append(AlertReportPart())
        return self.ReportPart[-1]


class EpisodicAlertReport(AbstractAlertReport):
    NODETYPE = msg.EpisodicAlertReport
    action = Actions.EpisodicAlertReport


class PeriodicAlertReport(AbstractAlertReport):
    NODETYPE = msg.PeriodicAlertReport
    action = Actions.PeriodicAlertReport


##### Component Report ###
class ComponentReportPart(ReportPartValuesList):
    ComponentState = cp.ContainerListProperty(
        msg.ComponentState,
        value_class=AbstractDeviceComponentStateContainer,
        cls_getter=get_state_container_class,
        ns_helper=default_ns_helper,
    )
    _props = ('ComponentState',)

    @property
    def values_list(self) -> list[Any]:
        return self.ComponentState


class AbstractComponentReport(AbstractReport):
    ReportPart = cp.SubElementListProperty(msg.ReportPart, value_class=ComponentReportPart)
    _props = ('ReportPart',)

    def add_report_part(self) -> ComponentReportPart:
        self.ReportPart.append(ComponentReportPart())
        return self.ReportPart[-1]


class EpisodicComponentReport(AbstractComponentReport):
    NODETYPE = msg.EpisodicComponentReport
    action = Actions.EpisodicComponentReport


class PeriodicComponentReport(AbstractComponentReport):
    NODETYPE = msg.PeriodicComponentReport
    action = Actions.PeriodicComponentReport


##### Operation Invoked Report ###
class InvocationInfo(PropertyBasedPMType):
    TransactionId = cp.NodeTextProperty(msg.TransactionId, UnsignedIntConverter, min_length=1)
    InvocationState = cp.NodeEnumTextProperty(msg.InvocationState, InvocationState)
    InvocationError = cp.NodeEnumTextProperty(msg.InvocationError, InvocationError, is_optional=True)
    InvocationErrorMessage = cp.SubElementListProperty(msg.InvocationErrorMessage, value_class=LocalizedText)
    _props = ('TransactionId', 'InvocationState', 'InvocationError', 'InvocationErrorMessage')

    def __repr__(self):
        if self.InvocationError:
            text = ', '.join([e.text for e in self.InvocationErrorMessage])
            return (
                f'{self.__class__.__name__}(TransactionId={self.TransactionId}, '
                f'InvocationState={self.InvocationState.value}, '
                f'InvocationError={self.InvocationError}), '
                f'InvocationErrorMessage={text}'
            )
        return f'{self.__class__.__name__}(TransactionId={self.TransactionId}, InvocationState={self.InvocationState})'

    def add_error_message(
        self,
        text: str,
        lang: str | None = None,
        ref: str | None = None,
        version: int | None = None,
        text_width: LocalizedTextWidth | None = None,
    ) -> None:
        self.InvocationErrorMessage.append(LocalizedText(text, lang, ref, version, text_width))


class OperationInvokedReportPart(AbstractReportPart):
    InvocationInfo = cp.SubElementProperty(
        msg.InvocationInfo, value_class=InvocationInfo, default_py_value=InvocationInfo(),
    )
    InvocationSource = cp.SubElementProperty(msg.InvocationSource, value_class=InstanceIdentifier)
    OperationHandleRef = cp.HandleRefAttributeProperty('OperationHandleRef', is_optional=False)
    OperationTarget = cp.HandleRefAttributeProperty('OperationTarget')
    _props = ('InvocationInfo', 'InvocationSource', 'OperationHandleRef', 'OperationTarget')


class OperationInvokedReport(AbstractReport):
    NODETYPE = msg.OperationInvokedReport
    action = Actions.OperationInvokedReport
    ReportPart = cp.SubElementListProperty(msg.ReportPart, value_class=OperationInvokedReportPart)
    _props = ('ReportPart',)

    def add_report_part(self) -> OperationInvokedReportPart:
        self.ReportPart.append(OperationInvokedReportPart())
        return self.ReportPart[-1]


class AbstractSetResponse(MessageType):
    Extension = cp.ExtensionNodeProperty(ext.Extension)
    InvocationInfo = cp.SubElementProperty(
        msg.InvocationInfo, value_class=InvocationInfo, default_py_value=InvocationInfo(),
    )
    MdibVersion = cp.VersionCounterAttributeProperty('MdibVersion', implied_py_value=0)
    SequenceId = cp.AnyURIAttributeProperty('SequenceId', is_optional=False)
    InstanceId = cp.UnsignedIntAttributeProperty('InstanceId')
    _props = ('Extension', 'InvocationInfo', 'MdibVersion', 'SequenceId', 'InstanceId')


class SetContextStateResponse(AbstractSetResponse):
    NODETYPE = msg.SetContextStateResponse
    action = Actions.SetContextStateResponse


class SetValueResponse(AbstractSetResponse):
    NODETYPE = msg.SetValueResponse
    action = Actions.SetValueResponse


class SetStringResponse(AbstractSetResponse):
    NODETYPE = msg.SetStringResponse
    action = Actions.SetStringResponse


class ActivateResponse(AbstractSetResponse):
    NODETYPE = msg.ActivateResponse
    action = Actions.ActivateResponse


class SetAlertStateResponse(AbstractSetResponse):
    NODETYPE = msg.SetAlertStateResponse
    action = Actions.SetAlertStateResponse


class SetComponentStateResponse(AbstractSetResponse):
    NODETYPE = msg.SetComponentStateResponse
    action = Actions.SetComponentStateResponse


class SetMetricStateResponse(AbstractSetResponse):
    NODETYPE = msg.SetMetricStateResponse
    action = Actions.SetMetricStateResponse


### WaveformStream ###
# no report parts in this report!
class WaveformStream(AbstractReport):
    NODETYPE = msg.WaveformStream
    action = Actions.Waveform
    State = cp.ContainerListProperty(
        msg.State,
        value_class=RealTimeSampleArrayMetricStateContainer,
        cls_getter=get_state_container_class,
        ns_helper=default_ns_helper,
    )
    _props = ('State',)


### DescriptionModificationReport ###
class DescriptionModificationReportPart(AbstractReportPart):
    Descriptor = cp.ContainerListProperty(
        msg.Descriptor,
        value_class=AbstractDescriptorContainer,
        cls_getter=get_descriptor_container_class,
        ns_helper=default_ns_helper,
    )
    State = cp.ContainerListProperty(
        msg.State,
        value_class=AbstractStateContainer,
        cls_getter=get_state_container_class,
        ns_helper=default_ns_helper,
    )
    ParentDescriptor = cp.HandleRefAttributeProperty('ParentDescriptor')
    ModificationType = cp.EnumAttributeProperty(
        'ModificationType', enum_cls=DescriptionModificationType, implied_py_value=DescriptionModificationType.UPDATE,
    )
    _props = ('Descriptor', 'State', 'ParentDescriptor', 'ModificationType')


class DescriptionModificationReport(AbstractReport):
    NODETYPE = msg.DescriptionModificationReport
    action = Actions.DescriptionModificationReport
    ReportPart = cp.SubElementListProperty(msg.ReportPart, value_class=DescriptionModificationReportPart)
    _props = ('ReportPart',)

    def add_report_part(self) -> DescriptionModificationReportPart:
        self.ReportPart.append(DescriptionModificationReportPart())
        return self.ReportPart[-1]

    @classmethod
    def from_node(cls, node: xml_utils.LxmlElement) -> DescriptionModificationReport:
        instance = super().from_node(node)
        # ser parent_handle members of descriptors
        for report_part in instance.ReportPart:
            for d in report_part.Descriptor:
                d.parent_handle = report_part.ParentDescriptor
        return instance


class SystemErrorReportPart(AbstractReportPart):
    ErrorCode = cp.SubElementProperty(msg.ErrorCode, value_class=pm_types.CodedValue)
    ErrorInfo = cp.SubElementListProperty(msg.ErrorInfo, value_class=pm_types.LocalizedText)
    _props = ('ErrorCode', 'ErrorInfo')


class SystemErrorReport(AbstractReport):
    NODETYPE = msg.SystemErrorReport
    action = Actions.SystemErrorReport
    ReportPart = cp.SubElementListProperty(msg.ReportPart, value_class=SystemErrorReportPart)
    _props = ('ReportPart',)


class AbstractGet(MessageType):
    pass


class AbstractSet(MessageType):
    OperationHandleRef = cp.NodeStringProperty(msg.OperationHandleRef)
    _props = ('OperationHandleRef',)

    @property
    def argument(self) -> Any:
        return None


class AbstractGetResponse(MessageType):
    MdibVersion = cp.IntegerAttributeProperty('MdibVersion', implied_py_value=0)
    SequenceId = cp.StringAttributeProperty('SequenceId')
    InstanceId = cp.IntegerAttributeProperty('InstanceId')
    _props = ('MdibVersion', 'SequenceId', 'InstanceId')

    def set_mdib_version_group(self, mdib_version_group: MdibVersionGroup) -> None:
        self.MdibVersion = mdib_version_group.mdib_version
        self.SequenceId = mdib_version_group.sequence_id
        self.InstanceId = mdib_version_group.instance_id


class Channel(PropertyBasedPMType):
    container = cp.ContainerProperty(
        None,
        value_class=ChannelDescriptorContainer,
        cls_getter=get_descriptor_container_class,
        ns_helper=default_ns_helper,
    )
    Metric = cp.ContainerListProperty(
        pm.Metric,
        value_class=AbstractMetricDescriptorContainer,
        cls_getter=get_descriptor_container_class,
        ns_helper=default_ns_helper,
    )
    _props = ('container', 'Metric')


class Vmd(PropertyBasedPMType):
    container = cp.ContainerProperty(
        None,
        value_class=VmdDescriptorContainer,
        cls_getter=get_descriptor_container_class,
        ns_helper=default_ns_helper,
    )
    Channel = cp.SubElementListProperty(pm.Channel, value_class=Channel)
    _props = ('container', 'Channel')


class Mds(PropertyBasedPMType):
    container = cp.ContainerProperty(
        None,
        value_class=MdsDescriptorContainer,
        cls_getter=get_descriptor_container_class,
        ns_helper=default_ns_helper,
    )
    Vmd = cp.SubElementListProperty(pm.Vmd, value_class=Vmd)
    _props = ('container', 'Vmd')


class MdDescription(PropertyBasedPMType):
    Mds = cp.SubElementListProperty(pm.Mds, value_class=Mds)
    _props = ('Mds',)


class MdState(PropertyBasedPMType):
    State = cp.ContainerListProperty(
        pm.State, value_class=AbstractStateContainer, cls_getter=get_state_container_class, ns_helper=default_ns_helper,
    )
    _props = ('State',)



class GetMdib(AbstractGet):
    NODETYPE = msg.GetMdib
    action = Actions.GetMdib


class GetMdibResponse(AbstractGetResponse):
    NODETYPE = msg.GetMdibResponse
    action = Actions.GetMdibResponse
    Mdib = cp.AnyEtreeNodeProperty(None)
    _props = ('Mdib',)


class GetMdDescription(AbstractGet):
    NODETYPE = msg.GetMdDescription
    action = Actions.GetMdDescription
    HandleRef = cp.SubElementHandleRefListProperty(msg.HandleRef)
    _props = ('HandleRef',)


class GetMdDescriptionResponse(AbstractGetResponse):
    NODETYPE = msg.GetMdDescriptionResponse
    action = Actions.GetMdDescriptionResponse
    MdDescription = cp.SubElementProperty(
        msg.MdDescription, value_class=MdDescription, default_py_value=MdDescription(),
    )
    _props = ('MdDescription',)


class GetMdState(AbstractGet):
    NODETYPE = msg.GetMdState
    action = Actions.GetMdState
    HandleRef = cp.SubElementHandleRefListProperty(msg.HandleRef)
    _props = ('HandleRef',)


class GetMdStateResponse(AbstractGetResponse):
    NODETYPE = msg.GetMdStateResponse
    action = Actions.GetMdStateResponse
    MdState = cp.SubElementProperty(msg.MdState, value_class=MdState, default_py_value=MdState())
    _props = ('MdState',)


class GetDescriptor(AbstractGet):
    NODETYPE = msg.GetDescriptor
    action = Actions.GetDescriptor
    HandleRef = cp.SubElementHandleRefListProperty(msg.HandleRef)
    _props = ('HandleRef',)


class GetDescriptorResponse(AbstractGetResponse):
    NODETYPE = msg.GetDescriptorResponse
    action = Actions.GetDescriptorResponse
    Descriptor = cp.ContainerListProperty(
        msg.Descriptor,
        value_class=AbstractDescriptorContainer,
        cls_getter=get_state_container_class,
        ns_helper=default_ns_helper,
    )
    _props = ('Descriptor',)


class GetContainmentTree(AbstractGet):
    NODETYPE = msg.GetContainmentTree
    action = Actions.GetContainmentTree
    HandleRef = cp.SubElementHandleRefListProperty(msg.HandleRef)
    _props = ('HandleRef',)


class GetContainmentTreeResponse(AbstractGetResponse):
    NODETYPE = msg.GetContainmentTreeResponse
    action = Actions.GetContainmentTreeResponse
    ContainmentTree = cp.SubElementListProperty(msg.ContainmentTree, value_class=ContainmentTree)
    _props = ('ContainmentTree',)


class GetContextStates(AbstractGet):
    NODETYPE = msg.GetContextStates
    action = Actions.GetContextStates
    HandleRef = cp.SubElementHandleRefListProperty(msg.HandleRef)
    _props = ('HandleRef',)


class GetContextStatesResponse(AbstractGetResponse):
    NODETYPE = msg.GetContextStatesResponse
    action = Actions.GetContextStatesResponse
    ContextState = cp.ContainerListProperty(
        msg.ContextState,
        value_class=AbstractContextStateContainer,
        cls_getter=get_state_container_class,
        ns_helper=default_ns_helper,
    )
    _props = ('ContextState',)


class GetContextStatesByIdentification(AbstractGet):
    NODETYPE = msg.GetContextStatesByIdentification
    action = Actions.GetContextStatesByIdentification
    Identification = cp.SubElementListProperty(msg.Identification, value_class=InstanceIdentifier)
    ContextType = cp.QNameAttributeProperty('ContextType')
    _props = ('HandleRef',)


class GetContextStatesByIdentificationResponse(AbstractGetResponse):
    NODETYPE = msg.GetContextStatesByIdentificationResponse
    action = Actions.GetContextStatesByIdentificationResponse
    ContextState = cp.ContainerListProperty(
        msg.ContextState,
        value_class=AbstractContextStateContainer,
        cls_getter=get_state_container_class,
        ns_helper=default_ns_helper,
    )
    _props = ('ContextState',)


class GetContextStatesByFilter(AbstractGet):
    NODETYPE = msg.GetContextStatesByFilter
    action = Actions.GetContextStatesByFilter
    Filter = cp.SubElementStringListProperty(msg.Filter)
    _props = ('HandleRef',)


class GetContextStatesByFilterResponse(AbstractGetResponse):
    NODETYPE = msg.GetContextStatesByFilterResponse
    action = Actions.GetContextStatesByFilterResponse
    ContextState = cp.ContainerListProperty(
        msg.ContextState,
        value_class=AbstractContextStateContainer,
        cls_getter=get_state_container_class,
        ns_helper=default_ns_helper,
    )
    _props = ('ContextState',)


class GetSupportedLanguages(AbstractGet):
    NODETYPE = msg.GetSupportedLanguages
    action = Actions.GetSupportedLanguages


class GetSupportedLanguagesResponse(AbstractGetResponse):
    NODETYPE = msg.GetSupportedLanguagesResponse
    action = Actions.GetSupportedLanguagesResponse
    Lang = cp.SubElementStringListProperty(msg.Lang)
    _props = ('Lang',)


class GetLocalizedText(AbstractGet):
    NODETYPE = msg.GetLocalizedText
    action = Actions.GetLocalizedText
    Ref = cp.SubElementHandleRefListProperty(msg.Ref)
    Version = cp.NodeIntProperty(msg.Version, is_optional=True)
    Lang = cp.SubElementStringListProperty(msg.Lang)
    TextWidth = cp.SubElementTextListProperty(msg.TextWidth, value_class=LocalizedTextWidth)
    NumberOfLines = cp.SubElementTextListProperty(msg.NumberOfLines, value_class=int)
    _props = ('Ref', 'Version', 'Lang', 'TextWidth', 'NumberOfLines')


class GetLocalizedTextResponse(AbstractGetResponse):
    NODETYPE = msg.GetLocalizedTextResponse
    action = Actions.GetLocalizedTextResponse
    Text = cp.SubElementListProperty(msg.Text, value_class=LocalizedText)
    _props = ('Text',)


class SetContextState(AbstractSet):
    NODETYPE = msg.SetContextState
    action = Actions.SetContextState
    ProposedContextState = cp.ContainerListProperty(
        msg.ProposedContextState,
        value_class=AbstractContextStateContainer,
        cls_getter=get_state_container_class,
        ns_helper=default_ns_helper,
    )
    _props = ('ProposedContextState',)

    @property
    def argument(self) -> Any:
        return self.ProposedContextState


class SetValue(AbstractSet):
    NODETYPE = msg.SetValue
    action = Actions.SetValue
    RequestedNumericValue = cp.NodeDecimalProperty(msg.RequestedNumericValue)
    _props = ('RequestedNumericValue',)

    @property
    def argument(self) -> Any:
        return self.RequestedNumericValue


class SetString(AbstractSet):
    NODETYPE = msg.SetString
    action = Actions.SetString
    RequestedStringValue = cp.NodeStringProperty(msg.RequestedStringValue)
    _props = ('RequestedStringValue',)

    @property
    def argument(self) -> Any:
        return self.RequestedStringValue


class SetAlertState(AbstractSet):
    NODETYPE = msg.SetAlertState
    action = Actions.SetAlertState
    ProposedAlertState = cp.ContainerProperty(
        msg.ProposedAlertState,
        value_class=AbstractAlertStateContainer,
        cls_getter=get_state_container_class,
        ns_helper=default_ns_helper,
    )
    _props = ('ProposedAlertState',)

    @property
    def argument(self) -> Any:
        return self.ProposedAlertState


class SetMetricState(AbstractSet):
    NODETYPE = msg.SetMetricState
    action = Actions.SetMetricState
    ProposedMetricState = cp.ContainerListProperty(
        msg.ProposedMetricState,
        value_class=AbstractMetricStateContainer,
        cls_getter=get_state_container_class,
        ns_helper=default_ns_helper,
    )
    _props = ('ProposedMetricState',)

    @property
    def argument(self) -> Any:
        return self.ProposedMetricState


class Argument(PropertyBasedPMType):
    ArgValue = cp.NodeStringProperty(msg.ArgValue)
    _props = ('ArgValue',)


class Activate(AbstractSet):
    NODETYPE = msg.Activate
    action = Actions.Activate
    Argument = cp.SubElementListProperty(msg.Argument, value_class=Argument)
    _props = ('Argument',)

    def add_argument(self, arg_value: Argument):
        arg = Argument()
        arg.ArgValue = arg_value.ArgValue
        self.Argument.append(arg)

    @property
    def argument(self) -> Any:
        return self.Argument


class SetComponentState(AbstractSet):
    NODETYPE = msg.SetComponentState
    action = Actions.SetComponentState
    ProposedComponentState = cp.ContainerListProperty(
        msg.ProposedComponentState,
        value_class=AbstractDeviceComponentStateContainer,
        cls_getter=get_state_container_class,
        ns_helper=default_ns_helper,
    )
    _props = ('ProposedComponentState',)

    @property
    def argument(self) -> Any:
        return self.ProposedComponentState



import itertools
import math
from typing import Callable

CurveGenerator = Callable[[float, float, int], list[float]]


def sinus(min_value: float, max_value: float, samples: int) -> list[float]:
    """Return a list of values for one sinus curve period."""
    delta = 2 * math.pi / samples
    _values = [math.sin(i * delta) for i in range(samples)]  # -1 ... +1
    return [(n + 1) / 2.0 * (max_value - min_value) + min_value for n in _values]  # min ... max


def sawtooth(min_value: float, max_value: float, samples: int) -> list[float]:
    """Return a list of values for one sawtooth curve period."""
    delta = (max_value - min_value) / float(samples)
    return [min_value + i * delta for i in range(samples)]


def triangle(min_value: float, max_value: float, samples: int) -> list[float]:
    """Return a list of values for one triangle curve period."""
    min_value = float(min_value)
    max_value = float(max_value)
    delta = (max_value - min_value) / float(samples) * 2
    samples_cnt = int(samples / 2)
    return [min_value + i * delta for i in range(samples_cnt)] + [max_value - i * delta for i in range(samples_cnt)]


class WaveformGeneratorBase:
    """Generator of infinite curve, data is provided by a curve generator."""

    def __init__(self,
                 values_generator: CurveGenerator,
                 min_value: float,
                 max_value: float,
                 waveform_period: float,
                 sample_period: float):
        if sample_period >= waveform_period:
            raise ValueError(
                f'Choose a waveformperiod > sampleperiod. currently have wp={waveform_period}, sp={sample_period}')
        if sample_period <= 0 or waveform_period <= 0:
            raise ValueError('no values <= 0 allowed for sample_period and waveform_period')
        self.sample_period = sample_period
        samples = int(waveform_period / sample_period)
        self._values = values_generator(min_value, max_value, samples)
        self._generator = itertools.cycle(self._values)

    def next_samples(self, count: int) -> list[float]:
        """Get next values from generator."""
        return [next(self._generator) for _ in range(count)]


class TriangleGenerator(WaveformGeneratorBase):
    """Generator of infinite triangle curve."""

    def __init__(self, min_value: float, max_value: float, waveform_period: float, sample_period: float):
        super().__init__(triangle, min_value, max_value, waveform_period, sample_period)


class SawtoothGenerator(WaveformGeneratorBase):
    """Generator of infinite saw tooth curve."""

    def __init__(self, min_value: float, max_value: float, waveform_period: float, sample_period: float):
        super().__init__(sawtooth, min_value, max_value, waveform_period, sample_period)


class SinusGenerator(WaveformGeneratorBase):
    """Generator of infinite sinus curve."""

    def __init__(self, min_value: float, max_value: float, waveform_period: float, sample_period: float):
        super().__init__(sinus, min_value, max_value, waveform_period, sample_period)



from __future__ import annotations

from decimal import Decimal
from typing import TYPE_CHECKING

from sdc11073.dispatch import DispatchKey
from sdc11073.namespaces import PrefixesEnum
from sdc11073.xml_types import msg_qnames
from sdc11073.xml_types.actions import Actions
from sdc11073.xml_types.addressing_types import HeaderInformationBlock

from .serviceclientbase import HostedServiceClient

if TYPE_CHECKING:
    from concurrent.futures import Future

    from sdc11073.consumer.manipulator import RequestManipulatorProtocol
    from sdc11073.mdib.statecontainers import AbstractStateProtocol
    from sdc11073.xml_types.msg_types import Argument


class SetServiceClient(HostedServiceClient):
    """Client for SetService."""

    port_type_name = PrefixesEnum.SDC.tag('SetService')
    notifications = (DispatchKey(Actions.OperationInvokedReport, msg_qnames.OperationInvokedReport),)

    def set_numeric_value(self, operation_handle: str,
                          requested_numeric_value: Decimal | float | int | str,
                          request_manipulator: RequestManipulatorProtocol | None = None) -> Future:
        """Send a GetSupportedLanguages request.

        :param operation_handle: a string
        :param requested_numeric_value: decimal, int, float or a string representing a decimal number
        :param request_manipulator:
        :return: a Future object
        """
        data_model = self._sdc_definitions.data_model
        self._logger.info('set_numeric_value operation_handle={} requested_numeric_value={}',  # noqa: PLE1205
                          operation_handle, requested_numeric_value)
        request = data_model.msg_types.SetValue()
        request.OperationHandleRef = operation_handle
        if isinstance(requested_numeric_value, float):
            # convert to string first in order to limit possible excessive number of digits
            request.RequestedNumericValue = Decimal(str(requested_numeric_value))
        else:
            request.RequestedNumericValue = Decimal(requested_numeric_value)
        inf = HeaderInformationBlock(action=request.action, addr_to=self.endpoint_reference.Address)
        message = self._msg_factory.mk_soap_message(inf, payload=request)
        return self._call_operation(message, request_manipulator=request_manipulator)

    def set_string(self, operation_handle: str,
                   requested_string: str,
                   request_manipulator: RequestManipulatorProtocol | None = None) -> Future:
        """Send a SetString request.

        :param operation_handle: a string
        :param requested_string: a string
        :param request_manipulator:
        :return: a Future object
        """
        data_model = self._sdc_definitions.data_model
        self._logger.info('set_string operation_handle={} requested_string={}',  # noqa: PLE1205
                          operation_handle, requested_string)
        request = data_model.msg_types.SetString()
        request.OperationHandleRef = operation_handle
        request.RequestedStringValue = requested_string
        inf = HeaderInformationBlock(action=request.action, addr_to=self.endpoint_reference.Address)
        message = self._msg_factory.mk_soap_message(inf, payload=request)
        return self._call_operation(message, request_manipulator=request_manipulator)

    def set_alert_state(self, operation_handle: str,
                        proposed_alert_state: AbstractStateProtocol,
                        request_manipulator: RequestManipulatorProtocol | None = None) -> Future:
        """Send a GetSupportedLanguages request.

        :param operation_handle: the handle of the operation to be called.
        :param proposed_alert_state: domainmodel.AbstractAlertState instance or a list of them
        :param request_manipulator:
        :return: a Future object
        """
        data_model = self._sdc_definitions.data_model
        self._logger.info('set_alert_state operation_handle={} requestedAlertState={}',  # noqa: PLE1205
                          operation_handle, proposed_alert_state)
        request = data_model.msg_types.SetAlertState()
        request.OperationHandleRef = operation_handle
        request.ProposedAlertState = proposed_alert_state
        inf = HeaderInformationBlock(action=request.action, addr_to=self.endpoint_reference.Address)
        message = self._msg_factory.mk_soap_message(inf, payload=request)
        return self._call_operation(message, request_manipulator=request_manipulator)

    def set_metric_state(self, operation_handle: str,
                         proposed_metric_states: list[AbstractStateProtocol],
                         request_manipulator: RequestManipulatorProtocol | None = None) -> Future:
        """Send a GetSupportedLanguages request.

        :param operation_handle: the handle of the operation to be called.
        :param proposed_metric_states: a list of domainmodel.AbstractMetricState instance or derived class
        :param request_manipulator:
        :return: a Future object
        """
        data_model = self._sdc_definitions.data_model
        self._logger.info('set_metric_state operation_handle={} requestedMetricState={}',  # noqa: PLE1205
                          operation_handle, proposed_metric_states)
        request = data_model.msg_types.SetMetricState()
        request.OperationHandleRef = operation_handle
        request.ProposedMetricState.extend(proposed_metric_states)
        inf = HeaderInformationBlock(action=request.action, addr_to=self.endpoint_reference.Address)
        message = self._msg_factory.mk_soap_message(inf, payload=request)
        return self._call_operation(message, request_manipulator=request_manipulator)

    def activate(self, operation_handle: str,
                 arguments: list[Argument] | None = None,
                 request_manipulator: RequestManipulatorProtocol | None = None) -> Future:
        """Send an Activate request.

        :param operation_handle: the handle of the operation to be called.
        :param arguments: a list of strings or None
        :param request_manipulator:
        :return: a concurrent.futures.Future object
        """
        data_model = self._sdc_definitions.data_model
        self._logger.info('activate handle={} arguments={}', operation_handle, arguments)  # noqa: PLE1205
        request = data_model.msg_types.Activate()
        request.OperationHandleRef = operation_handle
        if arguments is not None:
            for arg_value in arguments:
                request.add_argument(arg_value)
        inf = HeaderInformationBlock(action=request.action, addr_to=self.endpoint_reference.Address)
        message = self._msg_factory.mk_soap_message(inf, payload=request)
        return self._call_operation(message, request_manipulator=request_manipulator)

    def set_component_state(self, operation_handle: str,
                            proposed_component_states: list[AbstractStateProtocol],
                            request_manipulator: RequestManipulatorProtocol | None = None) -> Future:
        """Send an SetComponentState request.

        The set_component_state method corresponds to the SetComponentStateOperation objects in the MDIB
        and allows to insert or modify context states.
        """
        data_model = self._sdc_definitions.data_model
        tmp = ', '.join([f'{st.__class__.__name__} (DescriptorHandle={st.DescriptorHandle})'
                         for st in proposed_component_states])
        self._logger.info('set_component_state {}', tmp)  # noqa: PLE1205
        request = data_model.msg_types.SetComponentState()
        request.OperationHandleRef = operation_handle
        request.ProposedComponentState.extend(proposed_component_states)
        inf = HeaderInformationBlock(action=request.action, addr_to=self.endpoint_reference.Address)
        message = self._msg_factory.mk_soap_message(inf, payload=request)
        self._logger.debug('set_component_state sends {}', lambda: message.serialize(pretty=True))  # noqa: PLE1205
        return self._call_operation(message, request_manipulator=request_manipulator)



"""operation provider role."""

from __future__ import annotations

from typing import TYPE_CHECKING

from sdc11073 import xml_utils
from sdc11073.provider.operations import ActivateOperation, ExecuteParameters, ExecuteResult, OperationDefinitionBase
from sdc11073.roles import providerbase

if TYPE_CHECKING:
    from sdc11073.mdib.descriptorcontainers import (
        AbstractOperationDescriptorProtocol,
        ActivateOperationDescriptorContainer,
    )
    from sdc11073.roles.providerbase import OperationClassGetter
    from sdc11073.xml_types import msg_types
    from sdc11073.xml_types.pm_types import ActivateOperationDescriptorArgument


class OperationProvider(providerbase.ProviderRole):
    """Handle operations that work on operation states.

    Provier handling the operation calls. Used for plug-a-thon tests.
    """

    def _handle_plugathon_activate(self, params: ExecuteParameters) -> ExecuteResult:
        descriptor: ActivateOperationDescriptorContainer = params.operation_instance.descriptor_container
        if len(descriptor.Argument) != len(params.operation_request.argument):
            msg = f'Expected {len(descriptor.Argument)} arguments, got {len(params.operation_request.argument)}'
            raise ValueError(msg)

        for description, value in zip(descriptor.Argument, params.operation_request.argument):
            description: ActivateOperationDescriptorArgument
            value: msg_types.Argument
            # these are types from the plug-a-thon. add more if needed
            if description.Arg == xml_utils.QName('{http://www.w3.org/2001/XMLSchema}string'):
                pass  # value is already a string. no need to check for correct type
            elif description.Arg == xml_utils.QName('{http://www.w3.org/2001/XMLSchema}decimal'):
                float(value.ArgValue)  # just check if it's a valid decimal
            elif description.Arg == xml_utils.QName('{http://www.w3.org/2001/XMLSchema}anyURI'):
                pass  # TODO: check for type  # noqa: FIX002, TD002, TD003
            else:
                msg = f'{description.Arg} is not implemented'
                raise NotImplementedError(msg)

        # metric update has same mdib version as operation result. forbidden by sdpi
        with self._mdib.metric_state_transaction() as mgr:
            state = mgr.get_state(descriptor.OperationTarget)
            if not state.MetricValue:
                state.mk_metric_value()
            state.MetricValue.Value = ''.join(
                arg.ArgValue for arg in params.operation_request.argument
            )  # required for pat test 6f
        return ExecuteResult(
            params.operation_instance.operation_target_handle,
            self._mdib.data_model.msg_types.InvocationState.FINISHED,
        )

    def make_operation_instance(
        self,
        operation_descriptor_container: AbstractOperationDescriptorProtocol,
        operation_cls_getter: OperationClassGetter,
    ) -> OperationDefinitionBase | None:
        """Create an operation instance for the given descriptor container."""
        if operation_descriptor_container.Handle == 'activate_1.sco.mds_0':
            cls: type[ActivateOperation] = operation_cls_getter(operation_descriptor_container.NODETYPE)
            return cls(
                operation_descriptor_container.Handle,
                operation_descriptor_container.OperationTarget,
                operation_handler=self._handle_plugathon_activate,
            )
        return super().make_operation_instance(operation_descriptor_container, operation_cls_getter)



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

from lxml import etree

from sdc11073.exceptions import ApiUsageError
from sdc11073.namespaces import default_ns_helper as ns_hlp
from sdc11073.xml_types import xml_structure as struct
from sdc11073.xml_types.basetypes import ElementWithText, MessageType, XMLTypeBase

if TYPE_CHECKING:
    from sdc11073.xml_types.addressing_types import HeaderInformationBlock
    from sdc11073 import xml_utils

CHECK_NAMESPACES = False  # can be used to enable additional checks for too many namespaces or undefined namespaces


class SoapResponseError(Exception):
    """Exception raised when Response could not be processed."""

    def __init__(self, response_envelope: ReceivedSoapMessage):
        super().__init__()
        self.response_envelope = response_envelope


class ExtendedDocumentInvalid(etree.DocumentInvalid):
    """Exception for invalid document."""


class Soap12Envelope:
    """Soap12Envelope represents an outgoing soap envelope."""

    __slots__ = ('_header_nodes', '_payload_element', '_nsmap', 'header_info_block')

    def __init__(self, ns_map: dict | None = None):
        self._header_nodes = []
        self._payload_element = None
        if ns_map is None:
            self._nsmap = {}
        else:
            self._nsmap = ns_map
        for prefix in (ns_hlp.S12, ns_hlp.WSA):  # these are always needed
            self._nsmap[prefix.prefix] = prefix.namespace
        self.header_info_block = None

    def add_header_element(self, element: xml_utils.LxmlElement):
        """Add element to soap header."""
        self._header_nodes.append(element)

    def set_header_info_block(self, header_info_block: HeaderInformationBlock):
        """header_info_block contains data needed by ws-addressing."""
        self.header_info_block = header_info_block

    @property
    def payload_element(self) -> xml_utils.LxmlElement:
        """Get payload of soap envelope ( child node of Body element)."""
        return self._payload_element

    @payload_element.setter
    def payload_element(self, element: xml_utils.LxmlElement):
        if self._payload_element is not None:
            raise ApiUsageError('there can be only one body object')
        self._payload_element = element

    @property
    def nsmap(self) -> dict:
        """Get the namespace dictionary for this envelope."""
        return self._nsmap

    @property
    def header_nodes(self) -> list[xml_utils.LxmlElement]:
        """Get the list of header nodes."""
        return self._header_nodes


class ReceivedSoapMessage:
    """Represents a received soap envelope."""

    __slots__ = ('msg_node', 'msg_name', 'raw_data', 'header_info_block', '_doc_root', 'header_node', 'body_node')

    def __init__(self, xml_text: bytes, doc_root: xml_utils.LxmlElement):
        self.raw_data = xml_text
        self._doc_root: xml_utils.LxmlElement = doc_root
        self.header_node: xml_utils.LxmlElement = self._doc_root.find(ns_hlp.S12.tag('Header'))
        self.body_node: xml_utils.LxmlElement = self._doc_root.find(ns_hlp.S12.tag('Body'))
        self.header_info_block: HeaderInformationBlock | None = None
        try:
            self.msg_node = self.body_node[0]
            self.msg_name = etree.QName(self.msg_node.tag)
        except IndexError:  # body has no content, this can happen
            self.msg_node = None
            self.msg_name = None


# the following classes are named exactly like the types in soap_envelope.xsd schema, which looks weird sometimes.
class faultcodeEnum(Enum):  # noqa: N801
    """Fault codes."""

    DATAENC = ns_hlp.S12.tag('DataEncodingUnknown')
    MUSTUNSERSTAND = ns_hlp.S12.tag('MustUnderstand')
    RECEIVER = ns_hlp.S12.tag('Receiver')
    SENDER = ns_hlp.S12.tag('Sender')
    VERSION_MM = ns_hlp.S12.tag('VersionMismatch')


class reasontext(ElementWithText):  # noqa: N801
    """Text with language attribute."""

    lang: str = struct.StringAttributeProperty(ns_hlp.XML.tag('lang'), default_py_value='en-US')
    _props = ('lang',)


class faultreason(XMLTypeBase):  # noqa: N801
    """List of reasontext."""

    Text: list[reasontext] = struct.SubElementListProperty(ns_hlp.S12.tag('Text'), value_class=reasontext)
    _props = ('Text',)


class subcode(XMLTypeBase):  # noqa: N801
    """Value is a QName."""

    Value: etree.QName = struct.NodeTextQNameProperty(ns_hlp.S12.tag('Value'))
    # optional Subcode Element intentionally omitted, it is of type subcode => recursion, bad idea!
    _props = ('Value',)


class faultcode(XMLTypeBase):  # noqa: N801
    """Code wit subcode."""

    Value = struct.NodeEnumQNameProperty(ns_hlp.S12.tag('Value'), faultcodeEnum)
    Subcode = struct.SubElementProperty(ns_hlp.S12.tag('Subcode'), value_class=subcode, is_optional=True)
    _props = ('Value', 'Subcode')


class Fault(MessageType):
    """A Fault can be returned instead of a soap envelope."""

    NODETYPE = ns_hlp.S12.tag('Fault')
    action = f'{ns_hlp.WSA.namespace}/fault'
    Code = struct.SubElementProperty(ns_hlp.S12.tag('Code'), value_class=faultcode, default_py_value=faultcode())
    Reason = struct.SubElementProperty(ns_hlp.S12.tag('Reason'), value_class=faultreason,
                                       default_py_value=faultreason())
    Node = struct.AnyUriTextElement(ns_hlp.S12.tag('Node'), is_optional=True)
    Role = struct.AnyUriTextElement(ns_hlp.S12.tag('Role'), is_optional=True)
    # Schema says Detail is an "any" type. Here it is modelled as a string that becomes the text of the Detail node
    Detail = struct.NodeStringProperty(ns_hlp.S12.tag('Detail'), is_optional=True)
    _props = ('Code', 'Reason', 'Node', 'Role', 'Detail')
    additional_namespaces = (ns_hlp.XML, ns_hlp.WSE)

    def add_reason_text(self, text: str, lang: str = 'en-US'):
        """Add reason text to list."""
        txt = reasontext()
        txt.lang = lang
        txt.text = text
        self.Reason.Text.append(txt)

    def set_sub_code(self, sub_code: etree.QName):
        """Set sub code."""
        self.Code.Subcode = subcode()
        self.Code.Subcode.Value = sub_code



from __future__ import annotations

import copy
import inspect
from typing import Any
import math
from lxml import etree

from sdc11073 import observableproperties as properties
from sdc11073.namespaces import QN_TYPE, NamespaceHelper
from sdc11073 import xml_utils


class ContainerBase:
    """Common base class for descriptors and states."""

    NODETYPE: etree.QName = None  # overwrite in derived classes! This is the BICEPS Type.
    node = properties.ObservableProperty()
    is_state_container = False
    is_descriptor_container = False

    # every class with container properties must provide a list of property names.
    # this list is needed to create sub elements in a certain order.
    # rule is : elements are sorted from this root class to derived class. Last derived class comes last.
    # Initialization is from left to right
    # This is according to the inheritance in BICEPS xml schema

    def __init__(self):
        self.node = None  # set in update_from_node
        for _, cprop in self.sorted_container_properties():
            cprop.init_instance_data(self)

    def get_actual_value(self, attr_name: str) -> Any:
        """Ignore default value and implied value, e.g. return None if value is not present in xml."""
        return getattr(self.__class__, attr_name).get_actual_value(self)

    def mk_node(self,
                tag: etree.QName,
                ns_helper: NamespaceHelper,
                parent_node: xml_utils.LxmlElement | None = None,
                set_xsi_type: bool = False) -> xml_utils.LxmlElement:
        """Create an etree node from instance data.

        :param tag: tag of the newly created node
        :param ns_helper: namespaces.NamespaceHelper instance
        :param parent_node: optional parent node
        :param set_xsi_type: if True, adds Type attribute to node
        :return: etree node
        """
        ns_map = ns_helper.partial_map(ns_helper.PM,
                                       ns_helper.MSG,
                                       ns_helper.XSI)
        if parent_node is not None:
            node = etree.SubElement(parent_node, tag, nsmap=ns_map)
        else:
            node = etree.Element(tag, nsmap=ns_map)

        self.update_node(node, ns_helper, set_xsi_type)
        return node

    def update_node(self, node: xml_utils.LxmlElement,
                    ns_helper: NamespaceHelper,
                    set_xsi_type: bool = False) -> xml_utils.LxmlElement:
        """Update node with own data.

        :param node: node to be updated
        :param ns_helper: namespaces.NamespaceHelper instance
        :param set_xsi_type:if True, adds Type attribute to node
        :return: etree node
        """
        if set_xsi_type and self.NODETYPE is not None:
            node.set(QN_TYPE, ns_helper.doc_name_from_qname(self.NODETYPE))
        for _, prop in self.sorted_container_properties():
            prop.update_xml_value(self, node)
        return node

    def update_from_node(self, node: xml_utils.LxmlElement):
        """Update members from node."""
        for _, cprop in self.sorted_container_properties():
            cprop.update_from_node(self, node)
        self.node = node

    def _update_from_other(self, other_container: ContainerBase, skipped_properties: list[str] | None):
        """Update all ContainerProperties."""
        if skipped_properties is None:
            skipped_properties = []
        for prop_name, _ in self.sorted_container_properties():
            if prop_name not in skipped_properties:
                new_value = getattr(other_container, prop_name)
                setattr(self, prop_name, copy.copy(new_value))

    def mk_copy(self, copy_node: bool = False) -> ContainerBase:
        """Make a copy of self."""
        copied = copy.copy(self)
        if copy_node and self.node is not None:
            copied.node = xml_utils.copy_element(self.node)
        return copied

    def sorted_container_properties(self) -> list:
        """Return a list of (name, object) tuples of all GenericProperties ( and subclasses).

        Base class properties are first.
        """
        ret = []
        all_classes = inspect.getmro(self.__class__)
        for cls in reversed(all_classes):
            try:
                names = cls.__dict__['_props']
            except KeyError:
                continue
            for name in names:
                obj = getattr(cls, name)
                if obj is not None:
                    ret.append((name, obj))
        return ret

    def diff(self, other: ContainerBase,
             ignore_property_names: list[str] | None = None,
             max_float_diff = 1e-15) -> None | list[str]:
        """Compare all properties (except to be ignored ones).

        :param other: the object to compare with
        :param ignore_property_names: list of properties that shall be excluded from diff calculation
        :param max_float_diff: parameter for math.isclose() if float values are incorporated.
                                1e-15 corresponds to 15 digits max. accuracy (see sys.float_info.dig)
        :return: textual representation of differences or None if equal
        """
        ret = []
        ignore_list = ignore_property_names or []
        my_properties = self.sorted_container_properties()
        for name, _ in my_properties:
            if name in ignore_list:
                continue
            my_value = getattr(self, name)
            try:
                other_value = getattr(other, name)
            except AttributeError:
                ret.append(f'{name}={my_value}, other does not have this attribute')
            else:
                if isinstance(my_value, float) or isinstance(other_value, float):
                    if not math.isclose(my_value, other_value, rel_tol=max_float_diff, abs_tol=max_float_diff):
                        ret.append(f'{name}={my_value}, other={other_value}')
                elif my_value != other_value:
                    ret.append(f'{name}={my_value}, other={other_value}')
        # check also if other has a different list of properties
        my_property_names = {p[0] for p in my_properties}  # set comprehension
        other_property_names = {p[0] for p in other.sorted_container_properties()}
        surplus_names = other_property_names - my_property_names
        if surplus_names:
            ret.append(f'other has more data elements:{surplus_names}')
        return None if len(ret) == 0 else ret

    def is_equal(self, other: ContainerBase) -> bool:
        """Compare all properties."""
        return len(self.diff(other)) == 0



"""Implementation of metric provider functionality."""
from __future__ import annotations

from typing import TYPE_CHECKING, cast

from sdc11073.mdib.statecontainers import AbstractMetricStateContainer, MetricStateProtocol
from sdc11073.provider.operations import ExecuteResult
from sdc11073.xml_types.pm_types import ComponentActivation

from .providerbase import ProviderRole

if TYPE_CHECKING:
    from collections.abc import Iterable

    from sdc11073.mdib.descriptorcontainers import AbstractOperationDescriptorProtocol
    from sdc11073.mdib.mdibprotocol import ProviderMdibProtocol
    from sdc11073.mdib.transactionsprotocol import StateTransactionManagerProtocol, TransactionItem
    from sdc11073.provider.operations import ExecuteParameters, OperationDefinitionBase

    from .providerbase import OperationClassGetter


class GenericMetricProvider(ProviderRole):
    """Generic Handler.

    This is a generic Handler for
    - SetValueOperation on numeric metrics
    - SetStringOperation on (enum) string metrics
    """

    def __init__(self, mdib: ProviderMdibProtocol,
                 activation_state_can_remove_metric_value: bool = True,
                 log_prefix: str | None = None):
        """Create a GenericMetricProvider."""
        super().__init__(mdib, log_prefix)
        self.activation_state_can_remove_metric_value = activation_state_can_remove_metric_value

    def make_operation_instance(self,
                                operation_descriptor_container: AbstractOperationDescriptorProtocol,
                                operation_cls_getter: OperationClassGetter) -> OperationDefinitionBase | None:
        """Create an OperationDefinition for SetContextStateOperationDescriptor.

        Handle following cases:
        SetValueOperation, target = NumericMetricDescriptor:
          => handler = _set_numeric_value
        SetStringOperation, target = (Enum)StringMetricDescriptor:
          => handler = _set_string
        SetMetricStateOperationDescriptor, target = any subclass of AbstractMetricDescriptor:
          => handler = _set_metric_state
        """
        pm_names = self._mdib.data_model.pm_names
        operation_target_handle = operation_descriptor_container.OperationTarget
        op_target_entity = self._mdib.entities.by_handle(operation_target_handle)

        if operation_descriptor_container.NODETYPE == pm_names.SetValueOperationDescriptor:  # noqa: SIM300
            if op_target_entity.node_type == pm_names.NumericMetricDescriptor:
                op_cls = operation_cls_getter(pm_names.SetValueOperationDescriptor)
                return op_cls(operation_descriptor_container.Handle,
                              operation_target_handle,
                              self._set_numeric_value,
                              coded_value=operation_descriptor_container.Type)
            return None
        if operation_descriptor_container.NODETYPE == pm_names.SetStringOperationDescriptor:  # noqa: SIM300
            if op_target_entity.node_type in (pm_names.StringMetricDescriptor,
                                              pm_names.EnumStringMetricDescriptor):
                op_cls = operation_cls_getter(pm_names.SetStringOperationDescriptor)
                return op_cls(operation_descriptor_container.Handle,
                              operation_target_handle,
                              self._set_string,
                              coded_value=operation_descriptor_container.Type)
            return None
        if operation_descriptor_container.NODETYPE == pm_names.SetMetricStateOperationDescriptor:  # noqa: SIM300
            op_cls = operation_cls_getter(pm_names.SetMetricStateOperationDescriptor)
            return op_cls(operation_descriptor_container.Handle,
                          operation_target_handle,
                          self._set_metric_state,
                          coded_value=operation_descriptor_container.Type)
        return None

    def _set_metric_state(self, params: ExecuteParameters) -> ExecuteResult:
        """Handle SetMetricState calls (ExecuteHandler)."""
        proposed_states = params.operation_request.argument
        params.operation_instance.current_value = proposed_states

        with self._mdib.metric_state_transaction() as mgr:
            for proposed_state in proposed_states:
                target_entity = self._mdib.entities.by_handle(proposed_state.DescriptorHandle)
                if target_entity.state.is_metric_state:
                    self._logger.info('updating %s with proposed metric state', target_entity.state)
                    target_entity.state.update_from_other_container(proposed_state,
                                                      skipped_properties=['StateVersion', 'DescriptorVersion'])
                    mgr.write_entity(target_entity)
                else:
                    self._logger.warning('_set_metric_state operation: ignore invalid referenced type %s in operation',
                                         target_entity.state.NODETYPE)
        return ExecuteResult(params.operation_instance.operation_target_handle,
                             self._mdib.data_model.msg_types.InvocationState.FINISHED)

    def on_pre_commit(self, mdib: ProviderMdibProtocol,  # noqa: ARG002
                      transaction: StateTransactionManagerProtocol):
        """Set state.MetricValue to None if state.ActivationState requires this."""
        if not self.activation_state_can_remove_metric_value:
            return
        if transaction.metric_state_updates:
            self._handle_metrics_component_activation(transaction.metric_state_updates.values())
        if transaction.rt_sample_state_updates:
            self._handle_metrics_component_activation(transaction.rt_sample_state_updates.values())

    def _handle_metrics_component_activation(self, metric_state_updates: Iterable[TransactionItem]):
        """Check if MetricValue shall be removed."""
        for tr_item in metric_state_updates:
            new_state = cast(AbstractMetricStateContainer, tr_item.new)
            if new_state is None or not new_state.is_metric_state:
                continue
            # SF717: check if MetricValue shall be automatically removed
            if new_state.ActivationState in (ComponentActivation.OFF,
                                             ComponentActivation.SHUTDOWN,
                                             ComponentActivation.FAILURE):
                if new_state.MetricValue is not None:
                    # remove metric value
                    self._logger.info('%s: remove metric value because ActivationState="%s", handle="%s"',
                                      self.__class__.__name__, new_state.ActivationState, new_state.DescriptorHandle)
                    new_state.MetricValue = None

    def _set_numeric_value(self, params: ExecuteParameters) -> ExecuteResult:
        """Set a numerical metric value (ExecuteHandler)."""
        value = params.operation_request.argument
        pm_types = self._mdib.data_model.pm_types
        self._logger.info('set value of %s via %s from %r to %r',
                          params.operation_instance.operation_target_handle,
                          params.operation_instance.handle,
                          params.operation_instance.current_value, value)
        params.operation_instance.current_value = value
        entity = self._mdib.entities.by_handle(params.operation_instance.operation_target_handle)
        state = cast(MetricStateProtocol, entity.state)
        if state.MetricValue is None:
            state.mk_metric_value()
        state.MetricValue.Value = value
        # SF1823: For Metrics with the MetricCategory = Set|Preset that are being modified as a result of a
        # SetValue or SetString operation a Metric Provider shall set the MetricQuality / Validity = Vld.
        if entity.descriptor.MetricCategory in (pm_types.MetricCategory.SETTING,
                                                pm_types.MetricCategory.PRESETTING):
            state.MetricValue.Validity = pm_types.MeasurementValidity.VALID

        with self._mdib.metric_state_transaction() as mgr:
            mgr.write_entity(entity)
        return ExecuteResult(params.operation_instance.operation_target_handle,
                             self._mdib.data_model.msg_types.InvocationState.FINISHED)

    def _set_string(self, params: ExecuteParameters) -> ExecuteResult:
        """Set a string value (ExecuteHandler)."""
        value = params.operation_request.argument
        self._logger.info('set value %s from %s to %s',
                          params.operation_instance.operation_target_handle,
                          params.operation_instance.current_value, value)
        params.operation_instance.current_value = value
        entity = self._mdib.entities.by_handle(params.operation_instance.operation_target_handle)
        state = cast(MetricStateProtocol, entity.state)
        if state.MetricValue is None:
            state.mk_metric_value()
        state.MetricValue.Value = value

        with self._mdib.metric_state_transaction() as mgr:
            mgr.write_entity(entity)
        return ExecuteResult(params.operation_instance.operation_target_handle,
                             self._mdib.data_model.msg_types.InvocationState.FINISHED)



"""Message Reader implementation."""
from __future__ import annotations

import copy
import traceback
from collections import namedtuple
from dataclasses import dataclass
from io import BytesIO
from typing import TYPE_CHECKING

from lxml import etree

from sdc11073.exceptions import ValidationError
from sdc11073.namespaces import QN_TYPE, default_ns_helper, text_to_qname
from sdc11073.schema_resolver import mk_schema_validator
from sdc11073.xml_types.addressing_types import HeaderInformationBlock

from .soapenvelope import Fault, ReceivedSoapMessage, faultcodeEnum

if TYPE_CHECKING:
    from types import ModuleType

    from sdc11073 import xml_utils
    from sdc11073.definitions_base import BaseDefinitions
    from sdc11073.loghelper import LoggerAdapter
    from sdc11073.mdib.descriptorcontainers import AbstractDescriptorProtocol
    from sdc11073.mdib.statecontainers import AbstractStateProtocol
    from sdc11073.namespaces import PrefixNamespace


def validate_node(node: xml_utils.LxmlElement, xml_schema: etree.XMLSchema, logger: LoggerAdapter):
    """Let xml_schema instance validate the node."""
    try:
        xml_schema.assertValid(node)
    except etree.DocumentInvalid as ex:
        logger.warning(traceback.format_exc())
        logger.warning(etree.tostring(node, pretty_print=True).decode('utf-8'))
        fault = Fault()
        fault.Code.Value = faultcodeEnum.SENDER
        fault.set_sub_code(default_ns_helper.WSE.tag('InvalidMessage'))
        fault.add_reason_text(f'validation error: {ex}')

        raise ValidationError(reason='document invalid', soap_fault=fault) from ex


def _get_text(node: xml_utils.LxmlElement, q_name: etree.QName) -> str | None:
    if node is None:
        return None
    tmp = node.find(q_name)
    if tmp is None:
        return None
    return tmp.text


OperationRequest = namedtuple('OperationRequest', 'operation_handle argument') # noqa: PYI024

SubscriptionEndResult = namedtuple('SubscriptionEndResult', 'status_list reason_list reference_parameter_list') # noqa: PYI024


@dataclass
class MdibVersionGroupReader:
    """Groups the attributes that identify a mdib version."""

    mdib_version: int
    sequence_id: str
    instance_id: int | None

    @classmethod
    def from_node(cls, node: xml_utils.LxmlElement) -> MdibVersionGroupReader:
        """Construct from a node with version attributes."""
        mdib_version = int(node.get('MdibVersion', '0'))
        sequence_id = node.get('SequenceId')
        instance_id = node.get('InstanceId')
        if instance_id is not None:
            instance_id = int(instance_id)
        if sequence_id is None:
            raise ValueError('missing mandatory SequenceId attribute')
        return cls(mdib_version, sequence_id, instance_id)


@dataclass(frozen=True)
class ReceivedMessage:
    """ReceivedMessage contains all data of a received Message."""

    msg_reader: MessageReader
    p_msg: ReceivedSoapMessage
    action: str | None
    q_name: etree.QName
    mdib_version_group: MdibVersionGroupReader


class MessageReader:
    """MessageReader does all the conversions from DOM trees (body of SOAP messages) to MDIB objects."""

    def __init__(self, sdc_definitions: type[BaseDefinitions],
                 additional_schema_specs: list[PrefixNamespace] | None,
                 logger: LoggerAdapter,
                 validate: bool = True):
        self.schema_specs = [entry.value for entry in sdc_definitions.data_model.ns_helper.prefix_enum]
        if additional_schema_specs is not None:
            self.schema_specs.extend(additional_schema_specs)
        self._logger = logger
        self._data_model = sdc_definitions.data_model
        self.ns_hlp = sdc_definitions.data_model.ns_helper
        self._validate = validate
        self._xml_schema: etree.XMLSchema = mk_schema_validator(self.schema_specs, self.ns_hlp)

    @property
    def msg_names(self) -> ModuleType:
        """Return a module with all qualified names of the BICEPS message model."""
        return self._data_model.msg_names

    @property
    def pm_names(self) -> ModuleType:
        """Return a module with all qualified names of the BICEPS participant model."""
        return self._data_model.pm_names

    @property
    def pm_types(self) -> ModuleType:
        """Return a module with participant model types."""
        return self._data_model.pm_types

    @property
    def msg_types(self) -> ModuleType:
        """Return a module with message model types."""
        return self._data_model.msg_types

    def get_descriptor_container_class(self, qname: etree.QName) -> type[AbstractDescriptorProtocol]:
        """Get the class that represents a BICEPS descriptor entity with given QName."""
        return self._data_model.get_descriptor_container_class(qname)

    def get_state_container_class(self, qname: etree.QName) -> type[AbstractStateProtocol]:
        """Return the class that represents a BICEPS state entity with given QName."""
        return self._data_model.get_state_container_class(qname)

    def read_received_message(self, xml_text: bytes, validate: bool = True) -> ReceivedMessage:
        """Read complete message with addressing, message_id, payload,..."""
        parser = etree.ETCompatXMLParser(resolve_entities=False)
        try:
            doc_root = etree.fromstring(xml_text, parser=parser)
        except etree.XMLSyntaxError as ex:
            self._logger.warning('Error reading response ex=%r xml=%s', ex, xml_text.decode('utf-8'))
            raise
        if validate:
            self._validate_node(doc_root)

        message = ReceivedSoapMessage(xml_text, doc_root)
        if message.msg_node is not None and validate:
            self._validate_node(message.msg_node)
        message.header_info_block = HeaderInformationBlock.from_node(message.header_node)

        mdib_version_group = None
        if message.msg_node is not None:
            try:
                mdib_version_group = MdibVersionGroupReader.from_node(message.msg_node)
            except ValueError:
                mdib_version_group = None
        return ReceivedMessage(self, message, message.header_info_block.Action,
                               message.msg_name, mdib_version_group)

    def read_get_mdib_response(self, received_message_data: ReceivedMessage) -> tuple[
        list[AbstractDescriptorProtocol], list[AbstractStateProtocol]]:
        """Return list of all descriptors and states in mdib of received message."""
        mdib_node = received_message_data.p_msg.msg_node[0]
        return self.read_get_mdib_payload(mdib_node)

    def read_get_mdib_payload(self, mdib_node: xml_utils.LxmlElement) -> tuple[
        list[AbstractDescriptorProtocol], list[AbstractStateProtocol]]:
        """Return list of all descriptors and states in mdib."""
        descriptors = []
        states = []
        md_descr_node = mdib_node.find(self.pm_names.MdDescription)
        md_state_node = mdib_node.find(self.pm_names.MdState)
        if md_descr_node is not None:
            descriptors = self._read_md_description_node(md_descr_node)
        if md_state_node is not None:
            states = self._read_md_state_node(md_state_node)
        return descriptors, states

    def read_mdib_xml(self, xml_text: bytes) -> tuple[list[AbstractDescriptorProtocol], list[AbstractStateProtocol]]:
        """Return list of all descriptors and states in mdib."""
        payload = self.read_xml_text(xml_text)
        q_name = etree.QName(payload.tag)
        if q_name == self.msg_names.GetMdibResponse:
            return self.read_get_mdib_payload(payload[0])
        return self.read_get_mdib_payload(payload)

    def read_xml_text(self, xml_text: bytes) -> xml_utils.LxmlElement:
        """Parse imput, return a node."""
        parser = etree.ETCompatXMLParser(resolve_entities=False)
        try:
            node = etree.fromstring(xml_text, parser=parser)
        except Exception as ex:
            print(f'load error "{ex}" in "{xml_text}"')
            raise
        self._validate_node(node)
        return node

    def _read_md_description_node(self, md_description_node: xml_utils.LxmlElement) -> list[AbstractDescriptorProtocol]:
        descriptions = []

        def add_children(parent_node: xml_utils.LxmlElement):
            p_handle = parent_node.get('Handle')
            for child_node in parent_node:
                if child_node.get('Handle') is not None:
                    container = self._mk_descriptor_container_from_node(child_node, p_handle)
                    descriptions.append(container)
                    add_children(child_node)

        # iterate over tree, collect all handles of vmds, channels and metric descriptors
        all_mds = md_description_node.findall(self.pm_names.Mds)
        for mds_node in all_mds:
            mds = self._mk_descriptor_container_from_node(mds_node, None)
            descriptions.append(mds)
            add_children(mds_node)
        return descriptions

    def _read_md_state_node(self, md_state_node: xml_utils.LxmlElement) -> list[AbstractStateProtocol]:
        """Parse a GetMdStateResponse or the MdState part of GetMdibResponse."""
        state_containers = []
        all_state_nodes = md_state_node.findall(self.pm_names.State)
        for state_node in all_state_nodes:
            state_containers.append(self._mk_state_container_from_node(state_node)) # noqa: PERF401
        return state_containers

    def _mk_descriptor_container_from_node(self, node: xml_utils.LxmlElement,
                                           parent_handle: str | None) -> AbstractDescriptorProtocol:
        node_type = node.get(QN_TYPE)
        node_type = text_to_qname(node_type, node.nsmap) if node_type is not None else etree.QName(node.tag)
        descr_cls = self.get_descriptor_container_class(node_type)
        return descr_cls.from_node(node, parent_handle)

    def _mk_state_container_from_node(self, node: xml_utils.LxmlElement,
                                      forced_type: etree.QName | None = None) -> AbstractStateProtocol:
        """Create a state container from a node.

        :param node: an etree node
        :param forced_type: if given, the QName that shall be used for class instantiation instead of the data in node
        """
        if forced_type is not None:
            node_type = forced_type
        else:
            node_type = node.get(QN_TYPE)
            if node_type is not None:
                node_type = text_to_qname(node_type, node.nsmap)

        descriptor_container = None
        st_cls = self.get_state_container_class(node_type)
        if st_cls is None:
            msg = f'body type {node_type} is not known'
            raise ValueError(msg)

        if node.tag != self.pm_names.State:
            node = copy.copy(node)  # make a copy, do not modify the original report
            node.tag = self.pm_names.State
        state = st_cls(descriptor_container)
        state.update_from_node(node)
        state.node = node
        return state

    def _validate_node(self, node: xml_utils.LxmlElement):
        if self._validate:
            validate_node(node, self._xml_schema, self._logger)

    @staticmethod
    def read_wsdl(wsdl_text: bytes) -> etree.ElementTree:
        """Make am ElementTree instance."""
        return etree.parse(BytesIO(wsdl_text), parser=etree.ETCompatXMLParser(resolve_entities=False))



from __future__ import annotations

import time
from dataclasses import dataclass
from typing import TYPE_CHECKING, Callable, Protocol

from sdc11073 import loghelper
from sdc11073.dispatch.request import RequestData
from sdc11073.exceptions import InvalidActionError
from sdc11073.pysoap.msgfactory import CreatedMessage
from sdc11073.pysoap.soapenvelope import Fault, faultcodeEnum

if TYPE_CHECKING:
    from lxml import etree


@dataclass(frozen=True)
class DispatchKey:
    """Describes a message + action combination.

    The class is used to associate a handler to a soap message by action - message combination.
    """

    action: str
    message_tag: etree.QName | None

    def __repr__(self) -> str:
        return f'{self.__class__.__name__} action={self.action} msg={self.message_tag}'


OnPostHandler = Callable[[RequestData], CreatedMessage]

OnGetHandler = Callable[[RequestData], bytes]


class RequestHandlerProtocol(Protocol):
    """A request handler handles POST and GET requests."""

    def on_post(self, request_data: RequestData) -> CreatedMessage:
        """Dispatch POST requests."""

    def on_get(self, request_data: RequestData) -> str:
        """Dispatch GET requests."""


class RequestDispatcherProtocol(RequestHandlerProtocol):
    """The Protocol of a RequestDispatcher.

    A RequestDispatcher forwards messages to registered handlers. It implements the RequestHandlerProtocol.
    """

    def __init__(self, log_prefix: str):
        """Construct an instance."""

    def register_post_handler(self, dispatch_key: DispatchKey, on_post_handler: OnPostHandler):
        """Register a POST handler for a DispatchKey.

        The DispatchKey is a specific Soap message body and action string combination.
        An incoming message with this message body and action is forwarded to th on_post_handler.
        """

    def register_get_handler(self, dispatch_key: str, on_get_handler: OnGetHandler):
        """Register a GET handler for a DispatchKey.

        The DispatchKey is a specific Soap message body and action string combination.
        An incoming message with this message body and action is forwarded to th on_get_handler.
        """


class RequestDispatcher(RequestDispatcherProtocol):  # derive from protocol to help typing identify it.
    """The RequestDispatcher forwards messages to registered handlers.

    it implements the RequestDispatcherProtocol.
    """

    def __init__(self, log_prefix: str | None = None):
        self._post_handlers = {}
        self._get_handlers = {}
        self._logger = loghelper.get_logger_adapter(f'sdc.device.{self.__class__.__name__}', log_prefix)

    def register_post_handler(self, dispatch_key: DispatchKey, on_post_handler: OnPostHandler):
        """See documentation in RequestDispatcherProtocol."""
        self._post_handlers[dispatch_key] = on_post_handler

    def register_get_handler(self, dispatch_key: str, on_get_handler: OnGetHandler):
        """See documentation in RequestDispatcherProtocol."""
        self._get_handlers[dispatch_key] = on_get_handler

    def on_post(self, request_data: RequestData) -> CreatedMessage:
        """See documentation in RequestHandlerProtocol."""
        begin = time.monotonic()
        action = request_data.message_data.action
        func = self._get_post_handler(request_data)
        if func is None:
            fault = Fault()
            fault.Code.Value = faultcodeEnum.SENDER
            fault.add_reason_text(f'invalid action {action}')

            raise InvalidActionError(fault)
        returned_message = func(request_data)
        duration = time.monotonic() - begin
        self._logger.debug(  # noqa: PLE1205
            'incoming soap action "{}" to {}: duration={:.3f}sec.', action, request_data.path_elements, duration)
        return returned_message

    def on_get(self, request_data: RequestData) -> str:
        """See documentation in RequestHandlerProtocol."""
        begin = time.monotonic()
        key = request_data.current_path_element  # the current path element is the dispatch key
        func = self._get_handlers.get(key)
        if func is not None:
            self._logger.debug('on_get:path="{}" ,function="{}"', key, func.__name__)  # noqa: PLE1205
            result = func()
            duration = time.monotonic() - begin
            self._logger.debug('on_get:duration="{:.4f}"', duration)  # noqa: PLE1205
            return result
        error_text = f'on_get:path="{key}", no handler found!'
        self._logger.error(error_text)
        raise KeyError(error_text)

    def _get_post_handler(self, request_data: RequestData) -> OnPostHandler:
        key = DispatchKey(request_data.message_data.action, request_data.message_data.q_name)
        handler = self._post_handlers.get(key)
        if handler is None:
            self._logger.info('no handler for key={}', key)  # noqa: PLE1205
        return self._post_handlers.get(key)



from .porttypebase import DPWSPortTypeBase, WSDLMessageDescription, WSDLOperationBinding, mk_wsdl_two_way_operation
from .porttypebase import msg_prefix
from sdc11073.dispatch import DispatchKey
from sdc11073.namespaces import PrefixesEnum


class GetService(DPWSPortTypeBase):
    port_type_name = PrefixesEnum.SDC.tag('GetService')
    WSDLMessageDescriptions = (WSDLMessageDescription('GetMdState',
                                                      (f'{msg_prefix}:GetMdState',)),
                               WSDLMessageDescription('GetMdStateResponse',
                                                      (f'{msg_prefix}:GetMdStateResponse',)),
                               WSDLMessageDescription('GetMdib',
                                                      (f'{msg_prefix}:GetMdib',)),
                               WSDLMessageDescription('GetMdibResponse',
                                                      (f'{msg_prefix}:GetMdibResponse',)),
                               WSDLMessageDescription('GetMdDescription',
                                                      (f'{msg_prefix}:GetMdDescription',)),
                               WSDLMessageDescription('GetMdDescriptionResponse',
                                                      (f'{msg_prefix}:GetMdDescriptionResponse',)),
                               )
    WSDLOperationBindings = (WSDLOperationBinding('GetMdState', 'literal', 'literal'),
                             WSDLOperationBinding('GetMdib', 'literal', 'literal'),
                             WSDLOperationBinding('GetMdDescription', 'literal', 'literal'),)

    def register_hosting_service(self, hosting_service):
        super().register_hosting_service(hosting_service)
        actions = self._sdc_device.mdib.sdc_definitions.Actions
        msg_names = self._sdc_device.mdib.sdc_definitions.data_model.msg_names
        hosting_service.register_post_handler(DispatchKey(actions.GetMdState, msg_names.GetMdState),
                                              self._on_get_md_state)
        hosting_service.register_post_handler(DispatchKey(actions.GetMdib, msg_names.GetMdib),
                                              self._on_get_mdib)
        hosting_service.register_post_handler(DispatchKey(actions.GetMdDescription, msg_names.GetMdDescription),
                                              self._on_get_md_description)

    def _on_get_md_state(self, request_data):
        data_model = self._sdc_definitions.data_model
        msg_node = request_data.message_data.p_msg.msg_node
        get_md_state = data_model.msg_types.GetMdState.from_node(msg_node)
        requested_handles = get_md_state.HandleRef
        if len(requested_handles) > 0:
            self._logger.debug('_on_get_md_state from {} req. handles:{}', request_data.peer_name, requested_handles)
        else:
            self._logger.debug('_on_get_md_state from {}', request_data.peer_name)

        # get the requested state containers from mdib
        state_containers = []
        with self._mdib.mdib_lock:
            if len(requested_handles) == 0:
                # MessageModel: If the HANDLE reference list is empty, all states in the MDIB SHALL be included in the result list.
                state_containers.extend(self._mdib.states.objects)
                if self._sdc_device.contextstates_in_getmdib:
                    state_containers.extend(self._mdib.context_states.objects)
            else:
                if self._sdc_device.contextstates_in_getmdib:
                    for handle in requested_handles:
                        try:
                            # If a HANDLE reference does match a multi state HANDLE, the corresponding multi state SHALL be included in the result list
                            state_containers.append(self._mdib.context_states.handle.get_one(handle))
                        except (KeyError, ValueError):
                            # If a HANDLE reference does match a descriptor HANDLE, all states that belong to the corresponding descriptor SHALL be included in the result list
                            state_containers.extend(self._mdib.states.descriptor_handle.get(handle, []))
                            state_containers.extend(self._mdib.context_states.descriptor_handle.get(handle, []))
                else:
                    for handle in requested_handles:
                        state_containers.extend(self._mdib.states.descriptor_handle.get(handle, []))

                self._logger.debug('_on_get_md_state requested Handles:{} found {} states', requested_handles,
                                   len(state_containers))

        factory = self._sdc_device.msg_factory
        response = data_model.msg_types.GetMdStateResponse()
        response.MdState.State.extend(state_containers)
        response.set_mdib_version_group(self._mdib.mdib_version_group)
        created_message = factory.mk_reply_soap_message(request_data, response)
        self._logger.debug('_on_get_md_state returns {}',
                           lambda: created_message.serialize())
        return created_message

    def _on_get_mdib(self, request_data):
        self._logger.debug('_on_get_mdib')
        if self._sdc_device.contextstates_in_getmdib:
            mdib_node, mdib_version_group = self._mdib.reconstruct_mdib_with_context_states()
        else:
            mdib_node, mdib_version_group = self._mdib.reconstruct_mdib()
        response = self._data_model.msg_types.GetMdibResponse()
        response.set_mdib_version_group(mdib_version_group)
        response.Mdib = mdib_node
        response = self._sdc_device.msg_factory.mk_reply_soap_message(request_data, response)
        return response

    def _on_get_md_description(self, request_data):
        """
        MdDescription comprises the requested set of MDS descriptors. Which MDS descriptors are included depends on the msg:GetMdDescription/msg:HandleRef list:
        - If the HANDLE reference list is empty, all MDS descriptors SHALL be included in the result list.
        - If a HANDLE reference does match an MDS descriptor, it SHALL be included in the result list.
        - If a HANDLE reference does not match an MDS descriptor (any other descriptor), the MDS descriptor that is in the parent tree of the HANDLE reference SHOULD be included in the result list.
        """
        # currently this implementation only supports a single mds.
        # => if at least one handle matches any descriptor, the one mds is returned, otherwise empty payload
        data_model = self._sdc_definitions.data_model

        self._logger.debug('_on_get_md_description')
        msg_node = request_data.message_data.p_msg.msg_node
        get_md_state = data_model.msg_types.GetMdDescription.from_node(msg_node)
        requested_handles = get_md_state.HandleRef
        if len(requested_handles) > 0:
            self._logger.info('_on_get_md_description requested Handles:{}', requested_handles)
        response = self.mk_get_mddescription_response_message(
            request_data, self._sdc_device.mdib, requested_handles)
        self._logger.debug('_on_get_md_description returns {}',
                           lambda: response.serialize())
        return response

    def mk_get_mddescription_response_message(self, request_data, mdib, requested_handles):
        """For simplification reason this implementation returns either all descriptors or none."""
        return_all = len(requested_handles) == 0  # if we have handles, we need to check them
        dummy_response = self._sdc_definitions.data_model.msg_types.GetMdDescriptionResponse()
        dummy_response.set_mdib_version_group(mdib.mdib_version_group)
        response = self._sdc_device.msg_factory.mk_reply_soap_message(request_data, dummy_response)
        # now add to payload_element
        response_node = response.p_msg.payload_element
        for handle in requested_handles:
            # if at least one requested handle is valid, return all.
            if mdib.descriptions.handle.get_one(handle, allow_none=True) is not None:
                return_all = True
                break
        if return_all:
            md_description_node, mdib_version_group = mdib.reconstruct_md_description()
            # append all children of md_description_node to msg_names.MdDescription node in response
            response_node[0].extend(md_description_node[:])
        return response

    def add_wsdl_port_type(self, parent_node):
        """
        add wsdl:portType node to parent_node.
        xml looks like this:
        <wsdl:portType name="GetService" dpws:DiscoveryType="dt:ServiceProvider">
          <wsp:Policy>
            <dpws:Profile wsp:Optional="true"/>
          </wsp:Policy>
          <wsdl:operation name="GetMdState">
            <wsdl:input message="msg:GetMdState"/>
            <wsdl:output message="msg:GetMdStateResponse"/>
          </wsdl:operation>
          ...
        </wsdl:portType>
        :param parent_node:
        :return:
        """
        port_type = self._mk_port_type_node(parent_node)
        mk_wsdl_two_way_operation(port_type, operation_name='GetMdState')
        mk_wsdl_two_way_operation(port_type, operation_name='GetMdib')
        mk_wsdl_two_way_operation(port_type, operation_name='GetMdDescription')



"""The module contains the implementation of ProviderMdib with ProviderEntityGetter Protocol."""
from __future__ import annotations

import uuid
from collections import defaultdict
from contextlib import AbstractContextManager, contextmanager
from pathlib import Path
from threading import Lock
from typing import TYPE_CHECKING, Any, Callable

from sdc11073 import loghelper
from sdc11073.definitions_base import ProtocolsRegistry
from sdc11073.loghelper import LoggerAdapter
from sdc11073.observableproperties import ObservableProperty
from sdc11073.pysoap.msgreader import MessageReader

from . import mdibbase
from .providermdibxtra import ProviderMdibMethods
from .transactions import mk_transaction
from .transactionsprotocol import AnyTransactionManagerProtocol, TransactionType

if TYPE_CHECKING:
    from lxml.etree import QName

    from sdc11073.definitions_base import BaseDefinitions

    from .entityprotocol import ProviderEntityGetterProtocol
    from .transactionsprotocol import (
        ContextStateTransactionManagerProtocol,
        DescriptorTransactionManagerProtocol,
        StateTransactionManagerProtocol,
        TransactionResultProtocol,
    )

TransactionFactory = Callable[[mdibbase.MdibBase, TransactionType, LoggerAdapter],
                              AnyTransactionManagerProtocol]


class ProviderEntityGetter(mdibbase.EntityGetter):
    """Implementation of ProviderEntityGetterProtocol."""

    def new_entity(self,
            node_type: QName,
            handle: str,
            parent_handle: str | None) -> mdibbase.Entity | mdibbase.MultiStateEntity:
        """Create an entity."""
        if (handle in self._mdib.descriptions.handle
                or handle in self._mdib.context_states.handle
                # or handle in self._new_entities
        ):
            raise ValueError('Handle already exists')

        descr_cls = self._mdib.data_model.get_descriptor_container_class(node_type)
        descriptor_container = descr_cls(handle=handle, parent_handle=parent_handle)
        parent_descriptor = self._mdib.descriptions.handle.get_one(parent_handle)
        if self._mdib.data_model.pm_names.MdsDescriptor == parent_descriptor.NODETYPE:
            descriptor_container.set_source_mds(parent_descriptor.Handle)
        else:
            descriptor_container.set_source_mds(parent_descriptor.source_mds)

        if descriptor_container.is_context_descriptor:
            new_entity = mdibbase.MultiStateEntity(self._mdib, descriptor_container, [])
        else:
            state_cls = self._mdib.data_model.get_state_container_class(descriptor_container.STATE_QNAME)
            state = state_cls(descriptor_container)
            new_entity = mdibbase.Entity(self._mdib,descriptor_container, state)
        return new_entity


class ProviderMdib(mdibbase.MdibBase):
    """Device side implementation of a mdib.

    Do not modify containers directly, use transactions for that purpose.
    Transactions keep track of changes and initiate sending of update notifications to clients.
    """

    transaction: TransactionResultProtocol | None = ObservableProperty(fire_only_on_changed_value=False)
    rt_updates = ObservableProperty(fire_only_on_changed_value=False)  # different observable for performance

    def __init__(self,
                 sdc_definitions: type[BaseDefinitions] | None = None,
                 log_prefix: str | None = None,
                 extra_functionality: type | None = None,
                 transaction_factory: TransactionFactory | None = None,
                 ):
        """Construct a ProviderMdib.

        :param sdc_definitions: defaults to sdc11073.definitions_sdc.SdcV1Definitions
        :param log_prefix: a string
        :param extra_functionality: class for extra functionality, default is ProviderMdibMethods
        :param transaction_factory: optional alternative transactions factory.
        """
        if sdc_definitions is None:
            from sdc11073.definitions_sdc import SdcV1Definitions  # lazy import, needed to brake cyclic imports
            sdc_definitions = SdcV1Definitions
        super().__init__(sdc_definitions,
                         loghelper.get_logger_adapter('sdc.device.mdib', log_prefix))
        if extra_functionality is None:
            extra_functionality = ProviderMdibMethods
        self._xtra = extra_functionality(self)
        self._tr_lock = Lock()  # transaction lock

        self.sequence_id = uuid.uuid4().urn  # this uuid identifies this mdib instance

        self._annotators = {}
        self.current_transaction = None

        self.pre_commit_handler = None  # pre_commit_handler can modify transaction if needed before it is committed
        self.post_commit_handler = None  # post_commit_handler can modify mdib if needed after it is committed
        self._transaction_factory = transaction_factory or mk_transaction
        self._retrievability_episodic = []  # a list of handles
        self.retrievability_periodic = defaultdict(list)
        self.entities: ProviderEntityGetterProtocol = ProviderEntityGetter(self)

    @property
    def xtra(self) -> Any:
        """Give access to extended functionality."""
        return self._xtra

    @contextmanager
    def _transaction_manager(self, #  noqa: PLR0912, C901
                             transaction_type: TransactionType,
                             set_determination_time: bool = True) -> AbstractContextManager[
        AnyTransactionManagerProtocol]:
        """Start a transaction, return a new transaction manager."""
        with self._tr_lock, self.mdib_lock:
            try:
                self.current_transaction = self._transaction_factory(self, transaction_type, self.logger)
                yield self.current_transaction

                if callable(self.pre_commit_handler):
                    self.pre_commit_handler(self, self.current_transaction)
                if self.current_transaction.error:
                    self._logger.info('transaction_manager: transaction without updates!')
                else:
                    # update observables
                    transaction_result = self.current_transaction.process_transaction(set_determination_time)
                    self.transaction = transaction_result

                    if transaction_result.alert_updates:
                        self.alert_by_handle = {st.DescriptorHandle: st for st in transaction_result.alert_updates}
                    if transaction_result.comp_updates:
                        self.component_by_handle = {st.DescriptorHandle: st for st in transaction_result.comp_updates}
                    if transaction_result.ctxt_updates:
                        self.context_by_handle = {st.Handle: st for st in transaction_result.ctxt_updates}
                    if transaction_result.descr_created:
                        self.new_descriptors_by_handle = {descr.Handle: descr for descr
                                                          in transaction_result.descr_created}
                    if transaction_result.descr_deleted:
                        self.deleted_descriptors_by_handle = {descr.Handle: descr for descr
                                                              in transaction_result.descr_deleted}
                    if transaction_result.descr_updated:
                        self.updated_descriptors_by_handle = {descr.Handle: descr for descr
                                                              in transaction_result.descr_updated}
                    if transaction_result.metric_updates:
                        self.metrics_by_handle = {st.DescriptorHandle: st for st in transaction_result.metric_updates}
                    if transaction_result.op_updates:
                        self.operation_by_handle = {st.DescriptorHandle: st for st in transaction_result.op_updates}
                    if transaction_result.rt_updates:
                        self.waveform_by_handle = {st.DescriptorHandle: st for st in transaction_result.rt_updates}

                    if callable(self.post_commit_handler):
                        self.post_commit_handler(self, self.current_transaction)
            finally:
                self.current_transaction = None

    @contextmanager
    def context_state_transaction(self) -> AbstractContextManager[ContextStateTransactionManagerProtocol]:
        """Return a transaction for context state updates."""
        with self._transaction_manager(TransactionType.context) as mgr:
            yield mgr

    @contextmanager
    def alert_state_transaction(self, set_determination_time: bool = True) \
            -> AbstractContextManager[StateTransactionManagerProtocol]:
        """Return a transaction for alert state updates."""
        with self._transaction_manager(TransactionType.alert, set_determination_time) as mgr:
            yield mgr

    @contextmanager
    def metric_state_transaction(self, set_determination_time: bool = True) \
            -> AbstractContextManager[StateTransactionManagerProtocol]:
        """Return a transaction for metric state updates (not real time samples!)."""
        with self._transaction_manager(TransactionType.metric, set_determination_time) as mgr:
            yield mgr

    @contextmanager
    def rt_sample_state_transaction(self, set_determination_time: bool = False) \
            -> AbstractContextManager[StateTransactionManagerProtocol]:
        """Return a transaction for real time sample state updates."""
        with self._transaction_manager(TransactionType.rt_sample, set_determination_time) as mgr:
            yield mgr

    @contextmanager
    def component_state_transaction(self) -> AbstractContextManager[StateTransactionManagerProtocol]:
        """Return a transaction for component state updates."""
        with self._transaction_manager(TransactionType.component) as mgr:
            yield mgr

    @contextmanager
    def operational_state_transaction(self) -> AbstractContextManager[StateTransactionManagerProtocol]:
        """Return a transaction for operational state updates."""
        with self._transaction_manager(TransactionType.operational) as mgr:
            yield mgr

    @contextmanager
    def descriptor_transaction(self) -> AbstractContextManager[DescriptorTransactionManagerProtocol]:
        """Return a transaction for descriptor updates.

        This transaction also allows to handle the states that relate to the modified descriptors.
        """
        with self._transaction_manager(TransactionType.descriptor) as mgr:
            yield mgr

    @classmethod
    def from_mdib_file(cls,
                       path: str,
                       protocol_definition: type[BaseDefinitions] | None = None,
                       xml_reader_class: type[MessageReader] | None = MessageReader,
                       log_prefix: str | None = None) -> ProviderMdib:
        """Construct mdib from a file.

        :param path: the input file path for creating the mdib
        :param protocol_definition: an optional object derived from BaseDefinitions, forces usage of this definition
        :param xml_reader_class: class that is used to read mdib xml file
        :param log_prefix: a string or None
        :return: instance.
        """
        with Path(path).open('rb') as the_file:
            xml_text = the_file.read()
        return cls.from_string(xml_text,
                               protocol_definition,
                               xml_reader_class,
                               log_prefix)

    @classmethod
    def from_string(cls,
                    xml_text: bytes,
                    protocol_definition: type[BaseDefinitions] | None = None,
                    xml_reader_class: type[MessageReader] = MessageReader,
                    log_prefix: str | None = None) -> ProviderMdib:
        """Construct mdib from a string.

        :param xml_text: the input string for creating the mdib
        :param protocol_definition: an optional object derived from BaseDefinitions, forces usage of this definition
        :param xml_reader_class: class that is used to read mdib xml file
        :param log_prefix: a string or None
        :return: instance.
        """
        # get protocol definition that matches xml_text
        if protocol_definition is None:
            for definition_cls in ProtocolsRegistry.protocols:
                pm_namespace = definition_cls.data_model.ns_helper.PM.namespace.encode('utf-8')
                if pm_namespace in xml_text:
                    protocol_definition = definition_cls
                    break
        if protocol_definition is None:
            raise ValueError('cannot create instance, no known BICEPS schema version identified')
        mdib = cls(protocol_definition, log_prefix=log_prefix)

        xml_msg_reader = xml_reader_class(protocol_definition, None, mdib.logger)
        descriptor_containers, state_containers = xml_msg_reader.read_mdib_xml(xml_text)

        mdib.add_description_containers(descriptor_containers)
        mdib.add_state_containers(state_containers)
        mdib.xtra.mk_state_containers_for_all_descriptors()
        mdib.xtra.set_states_initial_values()
        mdib.xtra.update_retrievability_lists()
        mdib.xtra.set_all_source_mds()
        return mdib



from enum import Enum

from sdc11073.namespaces import default_ns_helper as ns_hlp

_ActionsNamespace = ns_hlp.SDC.namespace


class Actions(str, Enum):
    """Central definition of all action strings used in BICEPS."""

    OperationInvokedReport = _ActionsNamespace + '/SetService/OperationInvokedReport'
    EpisodicContextReport = _ActionsNamespace + '/ContextService/EpisodicContextReport'
    EpisodicMetricReport = _ActionsNamespace + '/StateEventService/EpisodicMetricReport'
    EpisodicOperationalStateReport = _ActionsNamespace + '/StateEventService/EpisodicOperationalStateReport'
    EpisodicAlertReport = _ActionsNamespace + '/StateEventService/EpisodicAlertReport'
    EpisodicComponentReport = _ActionsNamespace + '/StateEventService/EpisodicComponentReport'
    PeriodicContextReport = _ActionsNamespace + '/ContextService/PeriodicContextReport'
    PeriodicMetricReport = _ActionsNamespace + '/StateEventService/PeriodicMetricReport'
    PeriodicOperationalStateReport = _ActionsNamespace + '/StateEventService/PeriodicOperationalStateReport'
    PeriodicAlertReport = _ActionsNamespace + '/StateEventService/PeriodicAlertReport'
    PeriodicComponentReport = _ActionsNamespace + '/StateEventService/PeriodicComponentReport'
    SystemErrorReport = _ActionsNamespace + '/StateEventService/SystemErrorReport'
    Waveform = _ActionsNamespace + '/WaveformService/WaveformStream'
    DescriptionModificationReport = _ActionsNamespace + '/DescriptionEventService/DescriptionModificationReport'
    GetMdib = _ActionsNamespace + '/GetService/GetMdib'
    GetMdibResponse = _ActionsNamespace + '/GetService/GetMdibResponse'
    GetMdState = _ActionsNamespace + '/GetService/GetMdState'
    GetMdStateResponse = _ActionsNamespace + '/GetService/GetMdStateResponse'
    GetMdDescription = _ActionsNamespace + '/GetService/GetMdDescription'
    GetMdDescriptionResponse = _ActionsNamespace + '/GetService/GetMdDescriptionResponse'
    GetContextStates = _ActionsNamespace + '/ContextService/GetContextStates'
    GetContextStatesResponse = _ActionsNamespace + '/ContextService/GetContextStatesResponse'
    GetContextStatesByIdentification = _ActionsNamespace + '/ContextService/GetContextStatesByIdentification'
    GetContextStatesByIdentificationResponse = _ActionsNamespace + '/ContextService/GetContextStatesByIdentificationResponse'
    GetContextStatesByFilter = _ActionsNamespace + '/ContextService/GetContextStatesByFilter'
    GetContextStatesByFilterResponse = _ActionsNamespace + '/ContextService/GetContextStatesByFilterResponse'
    SetContextState = _ActionsNamespace + '/ContextService/SetContextState'
    SetContextStateResponse = _ActionsNamespace + '/ContextService/SetContextStateResponse'
    GetSupportedLanguages = _ActionsNamespace + '/LocalizationService/GetSupportedLanguages'
    GetSupportedLanguagesResponse = _ActionsNamespace + '/LocalizationService/GetSupportedLanguagesResponse'
    GetLocalizedText = _ActionsNamespace + '/LocalizationService/GetLocalizedText'
    GetLocalizedTextResponse = _ActionsNamespace + '/LocalizationService/GetLocalizedTextResponse'
    Activate = _ActionsNamespace + '/SetService/Activate'
    ActivateResponse = _ActionsNamespace + '/SetService/ActivateResponse'
    SetString = _ActionsNamespace + '/SetService/SetString'
    SetStringResponse = _ActionsNamespace + '/SetService/SetStringResponse'
    SetValue = _ActionsNamespace + '/SetService/SetValue'
    SetValueResponse = _ActionsNamespace + '/SetService/SetValueResponse'
    SetAlertState = _ActionsNamespace + '/SetService/SetAlertState'
    SetAlertStateResponse = _ActionsNamespace + '/SetService/SetAlertStateResponse'
    SetMetricState = _ActionsNamespace + '/SetService/SetMetricState'
    SetMetricStateResponse = _ActionsNamespace + '/SetService/SetMetricStateResponse'
    SetComponentState = _ActionsNamespace + '/SetService/SetComponentState'
    SetComponentStateResponse = _ActionsNamespace + '/SetService/SetComponentStateResponse'
    GetDescriptor = _ActionsNamespace + '/ContainmentTreeService/GetDescriptor'
    GetDescriptorResponse = _ActionsNamespace + '/ContainmentTreeService/GetDescriptorResponse'
    GetContainmentTree = _ActionsNamespace + '/ContainmentTreeService/GetContainmentTree'
    GetContainmentTreeResponse = _ActionsNamespace + '/ContainmentTreeService/GetContainmentTreeResponse'


# some sets of actions, useful when user wants to exclude some actions from subscriptions.
# these are the typical sets:
periodic_actions = {Actions.PeriodicContextReport,
                    Actions.PeriodicMetricReport,
                    Actions.PeriodicOperationalStateReport,
                    Actions.PeriodicAlertReport,
                    Actions.PeriodicComponentReport,
                    }

periodic_actions_and_system_error_report = set(periodic_actions).add(Actions.SystemErrorReport)



from __future__ import annotations

from typing import TYPE_CHECKING

from .definitions_base import AbstractDataModel, BaseDefinitions
from .mdib.descriptorcontainers import get_container_class as get_descriptor_container_class
from .mdib.statecontainers import get_container_class as get_state_container_class
from .namespaces import default_ns_helper as ns_hlp
from .xml_types import actions, msg_qnames, msg_types, pm_qnames, pm_types

if TYPE_CHECKING:
    from types import ModuleType

    from lxml import etree

    from .namespaces import NamespaceHelper


class V1Model(AbstractDataModel):
    """Data Model for SDC first version (assuming there will be successors in the future...)."""

    def __init__(self):
        super().__init__()
        self._ns_hlp = ns_hlp

    def get_descriptor_container_class(self, type_qname: etree.QName) -> type:
        """Get the class that represents a BICEPS descriptor entity with given QName."""
        return get_descriptor_container_class(type_qname)

    def get_state_container_class(self, type_qname: etree.QName) -> type:
        """Get the class that represents a BICEPS state entity with given QName."""
        return get_state_container_class(type_qname)

    @property
    def pm_types(self) -> ModuleType:
        """Get a module with participant model types."""
        return pm_types

    @property
    def pm_names(self) -> ModuleType:
        """Get a module with all qualified names of the BICEPS participant model."""
        return pm_qnames

    @property
    def msg_types(self) -> ModuleType:
        """Get a module with message model types."""
        return msg_types

    @property
    def msg_names(self) -> ModuleType:
        """Get a module with all qualified names of the BICEPS message model."""
        return msg_qnames

    @property
    def ns_helper(self) -> NamespaceHelper:
        """Gives access to a NamespaceHelper."""
        return self._ns_hlp


class SdcV1Definitions(BaseDefinitions):  # pylint: disable=invalid-name
    """Constants for SDC first version (assuming there will be successors in the future...)."""

    DpwsDeviceType = ns_hlp.DPWS.tag('Device')
    MedicalDeviceType = ns_hlp.MDPWS.tag('MedicalDevice')
    ActionsNamespace = ns_hlp.SDC.namespace
    PortTypeNamespace = ns_hlp.SDC.namespace
    MedicalDeviceTypesFilter = (DpwsDeviceType, MedicalDeviceType)
    Actions = actions.Actions
    data_model = V1Model()



from __future__ import annotations

from typing import TYPE_CHECKING

from sdc11073.namespaces import PrefixesEnum
from sdc11073.xml_types.addressing_types import HeaderInformationBlock

from .serviceclientbase import GetRequestResult, HostedServiceClient

if TYPE_CHECKING:
    from sdc11073.consumer.manipulator import RequestManipulatorProtocol
    from sdc11073.xml_types.pm_types import LocalizedTextWidth


class LocalizationServiceClient(HostedServiceClient):
    """Client for LocalizationService."""

    port_type_name = PrefixesEnum.SDC.tag('LocalizationService')

    def get_localized_texts(self, refs: list[str] | None = None,  # noqa: PLR0913
                            version: int | None = None,
                            langs: list[str] | None = None,
                            text_widths: list[LocalizedTextWidth] | None = None,
                            number_of_lines: list[int] | None = None,
                            request_manipulator: RequestManipulatorProtocol | None = None) -> GetRequestResult:
        """Send a GetLocalizedText request.

        :param refs: optional list of reference names of the texts that are requested.
        :param version: optional revision of the referenced text that is requested.
        :param langs: optional list of language identifiers.
        :param text_widths: optional list of LocalizedTextWidth enums
        :param number_of_lines: optional list of integers.
        :param request_manipulator: see documentation of RequestManipulatorProtocol
        """
        data_model = self._sdc_definitions.data_model
        request = data_model.msg_types.GetLocalizedText()
        if refs is not None:
            request.Ref.extend(refs)
        if version is not None:
            request.Version = version
        if langs is not None:
            request.Lang.extend(langs)
        if text_widths is not None:
            request.TextWidth.extend(text_widths)
        if number_of_lines is not None:
            request.NumberOfLines.extend(number_of_lines)
        inf = HeaderInformationBlock(action=request.action, addr_to=self.endpoint_reference.Address)
        message = self._msg_factory.mk_soap_message(inf, payload=request)
        received_message_data = self.post_message(message, request_manipulator=request_manipulator)
        cls = data_model.msg_types.GetLocalizedTextResponse
        result = cls.from_node(received_message_data.p_msg.msg_node)

        return GetRequestResult(received_message_data, result)

    def get_supported_languages(self,
                                request_manipulator: RequestManipulatorProtocol | None = None) -> GetRequestResult:
        """Send a GetSupportedLanguages request."""
        data_model = self._sdc_definitions.data_model
        request = data_model.msg_types.GetSupportedLanguages()
        inf = HeaderInformationBlock(action=request.action, addr_to=self.endpoint_reference.Address)
        message = self._msg_factory.mk_soap_message(inf, payload=request)
        received_message_data = self.post_message(message, request_manipulator=request_manipulator)
        cls = data_model.msg_types.GetSupportedLanguagesResponse
        result = cls.from_node(received_message_data.p_msg.msg_node)
        return GetRequestResult(received_message_data, result)



"""base subscription manager classes."""

from __future__ import annotations

import http.client
import socket
import time
import uuid
from collections import deque
from threading import Thread
from typing import TYPE_CHECKING, Any, Protocol
from urllib.parse import urlparse

from lxml import etree

from sdc11073 import loghelper, multikey, observableproperties, xml_utils
from sdc11073.etc import apply_map
from sdc11073.pysoap.soapclient import HTTPReturnCodeError, SoapClientProtocol
from sdc11073.pysoap.soapenvelope import Fault, faultcodeEnum
from sdc11073.xml_types import eventing_types as evt_types
from sdc11073.xml_types import isoduration
from sdc11073.xml_types.addressing_types import HeaderInformationBlock
from sdc11073.xml_types.basetypes import MessageType

if TYPE_CHECKING:
    import urllib.parse
    from collections.abc import Sequence
    from urllib.parse import SplitResult

    from sdc11073.definitions_base import BaseDefinitions
    from sdc11073.dispatch import RequestData
    from sdc11073.mdib.mdibbase import MdibVersionGroup
    from sdc11073.pysoap.msgfactory import CreatedMessage, MessageFactory
    from sdc11073.pysoap.soapclientpool import SoapClientPool

MAX_ROUNDTRIP_VALUES = 20


class RoundTripData:
    """Roundtrip data."""

    def __init__(self, values: Sequence[Any] | None, abs_max: int | None):
        if values:
            self.values = list(values)  # make a copy
            self.min = min(values)
            self.max = max(values)
            self.avg = sum(values) / len(values)
            self.abs_max = abs_max
        else:
            self.values = None
            self.min = None
            self.max = None
            self.avg = None
            self.abs_max = None

    def __repr__(self) -> str:
        return f'min={self.min:.4f} max={self.max:.4f} avg={self.avg:.4f} absmax={self.abs_max:.4f}'


def _mk_dispatch_identifier(reference_parameters: list, path_suffix: str) -> tuple[str | None, str | None]:
    # this is always our own reference parameter. We know that is has max. one element,
    # and the text is the identifier of the subscription
    if path_suffix == '':
        path_suffix = None
    if len(reference_parameters) > 0:
        return reference_parameters[0].text, path_suffix
    return None, path_suffix


class SubscriptionBase:
    """Subscription base."""

    MAX_NOTIFY_ERRORS = 1
    IDENT_TAG = etree.QName('http.local.com', 'MyDevIdentifier')

    def __init__(  # noqa: PLR0913
        self,
        mgr,  # noqa: ANN001
        subscribe_request: evt_types.Subscribe,
        accepted_encodings: list[str],
        base_urls: list[SplitResult],
        max_subscription_duration: int,
        soap_client_pool: SoapClientPool,
        msg_factory: MessageFactory,
        log_prefix: str,
    ):
        """Subscription base.

        :param mgr: the parent subscription manager
        :param subscribe_request:
        :param base_urls:
        :param max_subscription_duration:
        :param soap_client_pool: where to get the soap client for sending data
        :param msg_factory:
        """
        self._mgr = mgr  # seems to be unused
        self._logger = loghelper.get_logger_adapter('sdc.device.subscrMgr', log_prefix)
        self.base_urls = base_urls
        self._msg_factory = msg_factory
        self._soap_client_pool = soap_client_pool
        self.filter_type = subscribe_request.Filter
        if self.filter_type is None:
            msg = f'No filter provided for {self.__class__.__name__}'
            raise ValueError(msg)

        self.mode = subscribe_request.Delivery.Mode
        self.notify_to_address = subscribe_request.Delivery.NotifyTo.Address
        self.notify_to_url = urlparse(self.notify_to_address)
        self.notify_ref_params = subscribe_request.Delivery.NotifyTo.ReferenceParameters
        self.end_to_address = None
        self.end_to_ref_params = []
        if subscribe_request.EndTo is not None:
            self.end_to_address = subscribe_request.EndTo.Address
            self.end_to_ref_params = subscribe_request.EndTo.ReferenceParameters
            if self.end_to_address is not None:
                self._end_to_url = urlparse(self.end_to_address)
            else:
                self._end_to_url = None

        self.identifier_uuid = uuid.uuid4()
        self.reference_parameters = []  # default: no reference parameters
        self.path_suffix = None  # used for path based dispatching

        self._max_subscription_duration = max_subscription_duration
        self._started = None
        self._expire_seconds = None
        self.renew(subscribe_request.Expires)  # sets self._started and self._expire_seconds
        self._accepted_encodings = accepted_encodings

        self.notify_errors = 0
        self._is_closed = False
        self._is_connection_error = False
        self.last_roundtrip_times = deque(
            maxlen=MAX_ROUNDTRIP_VALUES,
        )  # a list of last n roundtrip times for notifications
        self.max_roundtrip_time = 0
        self.unsubscribed_at: float | None = None  # for housekeeping

    def set_reference_parameter(self):
        """Create a ReferenceParameters instance with a reference parameter."""
        reference_parameter = etree.Element(self.IDENT_TAG)
        reference_parameter.text = self.identifier_uuid.hex
        self.reference_parameters.append(reference_parameter)

    def renew(self, expires: float | None):
        """Renew a subscription."""
        self._started = time.monotonic()
        if expires:
            self._expire_seconds = min(expires, self._max_subscription_duration)
        else:
            self._expire_seconds = self._max_subscription_duration

    def matches(self, what: Any) -> bool:  # noqa: ARG002
        """Check if "what" parameter matches the filter criteria.

        The FilterType of eventing standard is very generic, it also has "any" types.
        THe type of "what" is determined by the Dialect parameter of FilterType.
        :param what: something that identifies a filter criteria of a notification message.
        :return:
        """
        return False

    def _get_soap_client(self) -> SoapClientProtocol:
        return self._soap_client_pool.get_soap_client(self.notify_to_url.netloc, self._accepted_encodings, self)

    @property
    def remaining_seconds(self) -> float:
        """Get the remaining seconds until expiry."""
        # do not use int, because it would round down to 0 even if the subscription is still valid, e.g. 0.5 seconds
        duration = round(self._expire_seconds - (time.monotonic() - self._started), 2)
        return max(duration, 0)

    @property
    def expire_string(self) -> str:
        """Get a duration string until invalid subscription."""
        return isoduration.duration_string(self.remaining_seconds)

    @property
    def has_delivery_failure(self) -> bool:
        """Determine whether this subscription has a delivery failure or not."""
        return self.notify_errors >= self.MAX_NOTIFY_ERRORS

    @property
    def has_connection_error(self) -> bool:
        """Determine whether this subscription has a connection error or not."""
        return self._is_connection_error

    @property
    def is_valid(self) -> bool:
        """Determine whether the subscription is valid or not."""
        if self._is_closed:
            return False
        return self.remaining_seconds > 0 and not self.has_delivery_failure

    def send_notification_end_message(
        self,
        code: str = 'SourceShuttingDown',
        reason: str = 'Event source going off line.',
    ):
        """Send a notification end message."""
        url = self.base_urls[0]
        my_addr = f'{url.scheme}:{url.netloc}/{url.path}'

        if not self.is_valid:
            return
        subscription_end = evt_types.SubscriptionEnd()
        subscription_end.SubscriptionManager.Address = my_addr
        subscription_end.SubscriptionManager.ReferenceParameters = self.reference_parameters
        subscription_end.Status = code
        subscription_end.add_reason(reason, 'en-US')
        inf = HeaderInformationBlock(
            action=subscription_end.action,
            addr_to=self.end_to_address or self.notify_to_address,
            reference_parameters=self.end_to_ref_params or self.notify_ref_params,
        )
        message = self._msg_factory.mk_soap_message(inf, payload=subscription_end)

        try:
            url = self._end_to_url or self.notify_to_url
            soap_client = self._get_soap_client()
            soap_client.post_message_to(url.path, message, msg='send_notification_end_message')
            self.notify_errors = 0
            self._is_connection_error = False
            self._is_closed = True
        except Exception as ex:  # noqa: BLE001
            # it does not matter that we could not send the message - end is end ;)
            self._logger.info('could not send subscription end message, error = {}', ex)  # noqa: PLE1205

    def close_by_subscription_manager(self) -> None:
        """Close subscription."""
        self._logger.info(  # noqa: PLE1205
            'close subscription id={} to {}',
            self.identifier_uuid,
            self.notify_to_address,
        )
        self._is_closed = True
        self._soap_client_pool.forget_usr(self.notify_to_url.netloc, self)

    def is_closed(self) -> bool:
        """Determine whether the subscription is closed or not."""
        return self._is_closed

    def __repr__(self) -> str:
        try:
            ref_ident = '<none>' if self.notify_ref_params is None else str(self.notify_ref_params)
        except TypeError:
            ref_ident = '<unknown>'
        return (
            f'{self.__class__.__name__}(notify_to={self.notify_to_address} ident={ref_ident}, '
            f'my identifier={self.identifier_uuid.hex}, expires={self.remaining_seconds})'
        )

    def get_roundtrip_stats(self) -> RoundTripData:
        """Get rountrip stats."""
        if len(self.last_roundtrip_times) > 0:
            return RoundTripData(self.last_roundtrip_times, self.max_roundtrip_time)
        return RoundTripData(None, None)

    def _mk_notification_message(
        self,
        header_info: HeaderInformationBlock,
        body_node: xml_utils.LxmlElement,
    ) -> CreatedMessage:
        return self._msg_factory.mk_soap_message_etree_payload(header_info, body_node)


class ActionBasedSubscription(SubscriptionBase):
    """Subscription for specific actions.

    Actions are a space separated list of strings in FilterType.text.
    """

    def __init__(self, *args, **kwargs):  # noqa: ANN002, ANN003
        # split the filter sting into separate action strings and keep them
        self.actions_filter: list[str] = []
        self._short_filter_names: list[str] = []  # helper for shorter log entries
        super().__init__(*args, **kwargs)
        if self.filter_type is not None:
            self.actions_filter.extend(self.filter_type.text.split())
            self._short_filter_names = [f.split('/')[-1] for f in self.actions_filter]

    def matches(self, what: str) -> bool:
        """Get what matches what.

        :param what: this must be a string
        :return: True if argument matches one of the strings in self.actions_filter, else False
        """
        action = what.strip()  # just to be sure there are no spaces
        return any(filter_string.endswith(action) for filter_string in self.actions_filter)

    def short_filter_names(self) -> list[str]:
        """Short filter names."""
        return self._short_filter_names

    def __repr__(self) -> str:
        try:
            ref_ident = '<none>' if self.notify_ref_params is None else str(self.notify_ref_params)
        except TypeError:
            ref_ident = '<unknown>'
        return (
            f'{self.__class__.__name__}(notify_to={self.notify_to_address} ident={ref_ident}, '
            f'my identifier={self.identifier_uuid.hex}, expires={self.remaining_seconds}, '
            f'filter={self._short_filter_names})'
        )


class SubscriptionManagerProtocol(Protocol):
    """Methods of a subscription manager."""

    def set_base_urls(self, base_urls: Sequence[urllib.parse.SplitResult]):
        """Own address, it must be sent to subscribers in subscribe responses."""
        ...

    def on_subscribe_request(self, request_data: RequestData) -> CreatedMessage:
        """Handle a subscribe request."""
        ...

    def on_unsubscribe_request(self, request_data: RequestData) -> CreatedMessage:
        """Handle an unsubscribe request."""
        ...

    def on_get_status_request(self, request_data: RequestData) -> CreatedMessage:
        """Handle a get status request."""
        ...

    def on_renew_request(self, request_data: RequestData) -> CreatedMessage:
        """Handle a renewal request."""
        ...

    def stop_all(self, send_subscription_end: bool):
        """Call when a device is shut down. It can send subscription end messages to all subscribers."""
        ...


class SubscriptionsManagerBase:
    """Base class for subscription manager."""

    DEFAULT_MAX_SUBSCR_DURATION = 7200  # max. possible duration of a subscription
    # observable has tuple(action, mdib_version_group, body_node)
    sent_to_subscribers = observableproperties.ObservableProperty(fire_only_on_changed_value=False)

    def __init__(
        self,
        sdc_definitions: BaseDefinitions,
        msg_factory: MessageFactory,
        soap_client_pool: SoapClientPool,
        max_subscription_duration: float | None = None,
        log_prefix: str | None = None,
    ):
        self.sdc_definitions = sdc_definitions
        self._msg_factory = msg_factory
        self._soap_client_pool: SoapClientPool = soap_client_pool
        self._logger = loghelper.get_logger_adapter('sdc.device.subscrMgr', log_prefix)
        self._max_subscription_duration = max_subscription_duration or self.DEFAULT_MAX_SUBSCR_DURATION
        self._subscriptions = multikey.MultiKeyLookup()
        self._subscriptions.add_index(
            'dispatch_identifier',
            multikey.UIndexDefinition(lambda obj: _mk_dispatch_identifier(obj.reference_parameters, obj.path_suffix)),
        )
        self._subscriptions.add_index('identifier', multikey.UIndexDefinition(lambda obj: obj.identifier_uuid.hex))
        self._subscriptions.add_index('netloc', multikey.IndexDefinition(lambda obj: obj.notify_to_url.netloc))
        self.base_urls: Sequence[urllib.parse.SplitResult] | None = None
        self._housekeeping_thread = Thread(target=self._do_housekeeping, name='housekeeping', daemon=True)
        self._run_housekeeping_thread = False
        self._housekeeping_thread.start()

    def set_base_urls(self, base_urls: Sequence[urllib.parse.SplitResult]):
        """Set base url.

        A subscription manager must know its own address, it must be sent to subscribers in subscribe responses.
        """
        self.base_urls = base_urls

    def _mk_subscription_instance(self, request_data: RequestData) -> SubscriptionBase:
        raise NotImplementedError

    def on_subscribe_request(self, request_data: RequestData) -> CreatedMessage:
        """Handle subscribe request."""
        subscription = self._mk_subscription_instance(request_data)
        with self._subscriptions.lock:
            self._subscriptions.add_object(subscription)
        self._logger.info('new {}', subscription)  # noqa: PLE1205
        return self._mk_subscribe_response_message(request_data, subscription, self.base_urls)

    def _mk_subscribe_response_message(
        self,
        request_data: RequestData,
        subscription: SubscriptionBase,
        base_urls: Sequence[urllib.parse.SplitResult],
    ) -> CreatedMessage:
        subscribe_response = evt_types.SubscribeResponse()

        path = '/'.join(request_data.consumed_path_elements)
        path_suffix = '' if subscription.path_suffix is None else f'/{subscription.path_suffix}'
        subscription_address = f'{base_urls[0].scheme}://{base_urls[0].netloc}/{path}{path_suffix}'
        subscribe_response.SubscriptionManager.Address = subscription_address
        subscribe_response.SubscriptionManager.ReferenceParameters = subscription.reference_parameters
        subscribe_response.Expires = subscription.remaining_seconds
        return self._msg_factory.mk_reply_soap_message(request_data, subscribe_response)

    def on_unsubscribe_request(self, request_data: RequestData) -> CreatedMessage:
        """Handle unsubscribe request."""
        subscription: SubscriptionBase = self._get_subscription_for_request(request_data)
        nsh = self.sdc_definitions.data_model.ns_helper
        if subscription is None:
            fault = Fault()
            fault.Code.Value = faultcodeEnum.RECEIVER
            fault.set_sub_code(nsh.WSE.tag('InvalidMessage'))
            fault.add_reason_text('unknown Subscription identifier')
            response = self._msg_factory.mk_reply_soap_message(request_data, fault, ns_map=[nsh.WSE])
        else:
            unsubscribe_response = evt_types.UnsubscribeResponse()
            response = self._msg_factory.mk_reply_soap_message(request_data, unsubscribe_response)
            subscription.unsubscribed_at = time.time()  # allow housekeeping to delete it delayed.
        return response

    def on_get_status_request(self, request_data: RequestData) -> CreatedMessage:
        """Handle get status request."""
        data_model = self.sdc_definitions.data_model
        nsh = data_model.ns_helper

        self._logger.debug('on_get_status_request {}', lambda: request_data.message_data.p_msg.raw_data)  # noqa: PLE1205
        subscription: SubscriptionBase = self._get_subscription_for_request(request_data)
        if subscription is None:
            fault = Fault()
            fault.Code.Value = faultcodeEnum.RECEIVER
            fault.set_sub_code(nsh.WSE.tag('InvalidMessage'))
            fault.add_reason_text('unknown Subscription identifier')
            response = self._msg_factory.mk_reply_soap_message(request_data, fault)
        else:
            get_status_response = evt_types.GetStatusResponse()
            get_status_response.Expires = subscription.remaining_seconds
            response = self._msg_factory.mk_reply_soap_message(request_data, get_status_response)
        return response

    def on_renew_request(self, request_data: RequestData) -> CreatedMessage:
        """Handle renew request."""
        data_model = self.sdc_definitions.data_model
        nsh = data_model.ns_helper
        renew = evt_types.Renew.from_node(request_data.message_data.p_msg.msg_node)
        expires = renew.Expires
        subscription: SubscriptionBase = self._get_subscription_for_request(request_data)
        if subscription is None:
            fault = Fault()
            fault.Code.Value = faultcodeEnum.RECEIVER
            fault.set_sub_code(nsh.WSE.tag('InvalidMessage'))
            fault.add_reason_text('unknown Subscription identifier')
            response = self._msg_factory.mk_reply_soap_message(request_data, fault)
        else:
            subscription.renew(expires)
            renew_response = evt_types.RenewResponse()
            renew_response.Expires = subscription.remaining_seconds
            response = self._msg_factory.mk_reply_soap_message(request_data, renew_response)
        return response

    def stop_all(self, send_subscription_end: bool):
        """Stop all subscriptions."""
        self._logger.info('stop_all called')
        # stop housekeeping thread first to get it out of the way
        self._logger.debug('stop housekeeping thread')
        self._run_housekeeping_thread = False
        self._housekeeping_thread.join()
        self._logger.debug('housekeeping thread stopped')
        self._logger.debug('end all subscriptions')
        self._end_all_subscriptions(send_subscription_end)

    def _end_all_subscriptions(self, send_subscription_end: bool):
        # async variant has a different implementation!
        with self._subscriptions.lock:
            if send_subscription_end:
                tmp = [s for s in self._subscriptions.objects if s.unsubscribed_at is None]
                apply_map(lambda subscription: subscription.send_notification_end_message(), tmp)
            apply_map(lambda subscription: subscription.close_by_subscription_manager(), self._subscriptions.objects)
            self._subscriptions.clear()

    def _get_subscription_for_request(self, request_data: RequestData) -> SubscriptionBase:
        reference_parameters = request_data.message_data.p_msg.header_info_block.reference_parameters
        path_suffix = '/'.join(request_data.path_elements)  # not consumed path elements
        dispatch_identifier = _mk_dispatch_identifier(reference_parameters, path_suffix)
        with self._subscriptions.lock:
            subscription = self._subscriptions.dispatch_identifier.get_one(dispatch_identifier, allow_none=True)
        if subscription is None:
            self._logger.warning(  # noqa: PLE1205
                '{}: unknown Subscription identifier "{}" from {}',
                request_data.message_data.q_name,
                dispatch_identifier,
                request_data.peer_name,
            )
        return subscription

    def send_to_subscribers(
        self,
        payload: MessageType | xml_utils.LxmlElement,
        action: str,
        mdib_version_group: MdibVersionGroup,
    ):
        """Send a payload to the subscribers."""
        subscribers = self._get_subscriptions_for_action(action)
        nsh = self.sdc_definitions.data_model.ns_helper
        # convert to element tree only once for all subscribers
        if isinstance(payload, MessageType):
            namespaces = [nsh.PM, nsh.MSG]
            namespaces.extend(payload.additional_namespaces)
            my_ns_map = nsh.partial_map(*namespaces)
            body_node = payload.as_etree_node(payload.NODETYPE, my_ns_map)
        else:
            body_node = payload

        self.sent_to_subscribers = (action, mdib_version_group, body_node)  # update observable
        for subscriber in subscribers:
            self._logger.debug('{}: sending report to {}', action, subscriber.notify_to_address)  # noqa: PLE1205
            self._send_notification_report(subscriber, body_node, action)

    def _send_notification_report(self, subscription, body_node: xml_utils.LxmlElement, action: str):  # noqa: ANN001
        try:
            subscription.send_notification_report(body_node, action)
        except ConnectionRefusedError as ex:
            self._logger.error('could not send notification report: {!r}:  subscr = {}', ex, subscription)  # noqa: PLE1205, TRY400
        except HTTPReturnCodeError as ex:
            # this is an error related to the connection => log error and continue
            self._logger.error(  # noqa: PLE1205, TRY400
                'could not send notification report: HTTP status= {}, reason={}, {}',
                ex.status,
                ex.reason,
                subscription,
            )
        except http.client.NotConnected as ex:
            # this is an error related to the connection => log error and continue
            self._logger.error('could not send notification report: {!r}:  subscr = {}', ex, subscription)  # noqa: PLE1205, TRY400
        except socket.timeout as ex:
            # this is an error related to the connection => log error and continue
            self._logger.error('could not send notification report error= {!r}: {}', ex, subscription)  # noqa: PLE1205, TRY400
        except etree.DocumentInvalid as ex:
            # this is an error related to the document, it cannot be sent to any subscriber => re-raise
            self._logger.error('Invalid Document: {!r}\n{}', ex, etree.tostring(body_node))  # noqa: PLE1205, TRY400
            raise
        except Exception:
            # this should never happen! => re-raise
            self._logger.exception('could not send notification report for subscription: {}', subscription)  # noqa: PLE1205
            raise

    def _get_subscriptions_for_action(self, action: str) -> list[Any]:
        with self._subscriptions.lock:
            return [s for s in self._subscriptions.objects if s.matches(action)]

    def _do_housekeeping(self):
        """Remove expired or invalid subscriptions. Method is executed in a thread."""
        self._run_housekeeping_thread = True
        while self._run_housekeeping_thread:
            time.sleep(1)
            now = time.time()
            with self._subscriptions.lock:
                obsolete_subscriptions = [
                    s
                    for s in self._subscriptions.objects
                    if not s.is_valid or (s.unsubscribed_at is not None and now > s.unsubscribed_at + 1)
                ]

                for obsolete_subscription in obsolete_subscriptions:
                    if not obsolete_subscription.is_closed():
                        obsolete_subscription.close_by_subscription_manager()

                        self._subscriptions.remove_object(obsolete_subscription)
